from __future__ import annotations

import torch
import triton
import triton.language as tl
from fouroversix.utils import DataType, RoundStyle, ScaleRule
from triton.tools.tensor_descriptor import TensorDescriptor

E2M1_MAX_VALUE = tl.constexpr(6)
E2M1_MAX_FOUR = tl.constexpr(4)
E4M3_MAX_VALUE = tl.constexpr(448)
E4M3_MAX_FOUROVERSIX = tl.constexpr(256)
SCALE_MEGABLOCK_SIZE = tl.constexpr(512)

DATA_TYPE_MXFP4 = tl.constexpr(DataType.mxfp4.value)
DATA_TYPE_NVFP4 = tl.constexpr(DataType.nvfp4.value)

ROUND_STYLE_NEAREST = tl.constexpr(RoundStyle.nearest.value)
ROUND_STYLE_STOCHASTIC = tl.constexpr(RoundStyle.stochastic.value)

SCALE_RULE_ABS_MAX = tl.constexpr(ScaleRule.abs_max.value)
SCALE_RULE_STATIC_4 = tl.constexpr(ScaleRule.static_4.value)
SCALE_RULE_STATIC_6 = tl.constexpr(ScaleRule.static_6.value)
SCALE_RULE_MAE = tl.constexpr(ScaleRule.mae.value)
SCALE_RULE_MSE = tl.constexpr(ScaleRule.mse.value)


@triton.jit
def rht_kernel(
    x_desc,
    h_desc,
    y_desc,
    # Meta-parameters
    # TODO(jack): Update RHT kernel to support unpadded dimensions
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    TRANSPOSE: tl.constexpr,
) -> None:
    HAD_BLOCK_SIZE: tl.constexpr = h_desc.block_shape[0]

    pid_m = tl.program_id(0)
    pid_n = tl.program_id(1)

    # Load H [B, B]
    h_block = h_desc.load([0, 0])

    m_block_offset = pid_m * BLOCK_SIZE_M
    n_block_offset = pid_n * BLOCK_SIZE_N

    if not TRANSPOSE:
        x_block = x_desc.load([m_block_offset, n_block_offset])
    else:
        x_block = x_desc.load([n_block_offset, m_block_offset]).T

    y_block = tl.dot(
        x_block.reshape(
            BLOCK_SIZE_M * BLOCK_SIZE_N // HAD_BLOCK_SIZE,
            HAD_BLOCK_SIZE,
        ).to(tl.bfloat16),
        h_block,
    ).reshape(BLOCK_SIZE_M, BLOCK_SIZE_N)

    y_desc.store([m_block_offset, n_block_offset], y_block)


@triton.jit
def block_scaled_fp4_quantization_kernel(
    x_block,
    x_amax_ptr,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    FP4_FORMAT: tl.constexpr,
    ROUND_STYLE: tl.constexpr,
    BLOCK_SCALE_2D: tl.constexpr,
    SCALE_RULE: tl.constexpr,
    RBITS: tl.constexpr,
) -> None:
    E2M1_MAX_ALLOWED_VALUE: tl.constexpr = (
        E2M1_MAX_VALUE if SCALE_RULE == SCALE_RULE_STATIC_6 else E2M1_MAX_FOUR
    )

    if FP4_FORMAT == DATA_TYPE_MXFP4:
        x_scale_blocks = x_block.reshape(128, 4, 32)
        x_scales_hp = tl.max(x_scale_blocks.abs(), axis=-1) / E2M1_MAX_ALLOWED_VALUE
        x_scales_e8m0_u32 = x_scales_hp.cast(tl.uint32, bitcast=True)

        # Use the 8-bit exponent as the scale factor
        x_scales_e8m0 = ((x_scales_e8m0_u32 >> 23) & 0xFF).to(tl.uint8)

        # Add one in order to round up
        x_scales = tl.where(
            (x_scales_e8m0_u32 & 0x7FFFFF) == 0,
            x_scales_e8m0,
            x_scales_e8m0 + 1,
        )

        # Convert the rounded-up scale factor back to a 32-bit float
        x_scales_hp = (x_scales.cast(tl.uint32) << 23).cast(x_block.dtype, bitcast=True)

        if BLOCK_SCALE_2D:
            x_scales_hp = (
                tl.max(
                    x_scales_hp.reshape(4, 32, 4).permute(0, 2, 1),
                    axis=-1,
                )
                .expand_dims(0)
                .broadcast_to(4, 32, 4)
                .permute(1, 0, 2)
                .reshape(128, 4)
            )

        (x_block_scaled_b1, x_block_scaled_b2) = (
            (x_scale_blocks / x_scales_hp.expand_dims(2))
            .reshape(BLOCK_SIZE_M, BLOCK_SIZE_N // 2, 2)
            .split()
        )
    elif FP4_FORMAT == DATA_TYPE_NVFP4:
        x_amax = tl.load(x_amax_ptr)
        x_scale_blocks = x_block.reshape(128, 4, 16)

        if x_amax == 0:
            x_scales_hp = tl.full((128, 4), 0, dtype=tl.float32)
        else:
            encode_scale = tl.div_rn(E2M1_MAX_ALLOWED_VALUE * E4M3_MAX_VALUE, x_amax)
            x_scales_hp = (
                tl.div_rn(tl.max(x_scale_blocks.abs(), axis=-1), E2M1_MAX_ALLOWED_VALUE)
                * encode_scale
            )

        if BLOCK_SCALE_2D:
            x_scales_hp = (
                tl.max(
                    x_scales_hp.reshape(8, 16, 4).permute(0, 2, 1),
                    axis=-1,
                )
                .expand_dims(0)
                .broadcast_to(16, 8, 4)
                .permute(1, 0, 2)
                .reshape(128, 4)
            )

        x_scales = x_scales_hp.to(tl.float8e4nv)

        decode_scale = tl.div_rn(
            1,
            tl.div_rn(E2M1_MAX_ALLOWED_VALUE * E4M3_MAX_VALUE, x_amax),
        )
        (x_block_scaled_b1, x_block_scaled_b2) = (
            tl.where(
                x_scales.expand_dims(2).to(x_amax.dtype) != 0,
                x_scale_blocks
                * tl.div_rn(1, decode_scale * x_scales.to(x_amax.dtype).expand_dims(2)),
                0,
            )
            .reshape(BLOCK_SIZE_M, BLOCK_SIZE_N // 2, 2)
            .split()
        )

    if ROUND_STYLE == ROUND_STYLE_NEAREST:
        x_e2m1 = tl.inline_asm_elementwise(
            asm="""
                {
                .reg .b8 byte0, byte1, byte2, byte3;
                cvt.rn.satfinite.e2m1x2.f32 byte0, $5, $1;
                cvt.rn.satfinite.e2m1x2.f32 byte1, $6, $2;
                cvt.rn.satfinite.e2m1x2.f32 byte2, $7, $3;
                cvt.rn.satfinite.e2m1x2.f32 byte3, $8, $4;
                mov.b32 $0, {byte0, byte1, byte2, byte3};
                }
                """,
            constraints="=r,r,r,r,r,r,r,r,r",
            args=[x_block_scaled_b1, x_block_scaled_b2],
            dtype=tl.uint8,
            is_pure=True,
            pack=4,
        )
    elif ROUND_STYLE == ROUND_STYLE_STOCHASTIC:
        if RBITS == -1:
            rbits = tl.rand(
                0,
                tl.arange(0, BLOCK_SIZE_M)[:, None] * BLOCK_SIZE_N // 2
                + tl.arange(0, BLOCK_SIZE_N // 2)[None, :],
            ).cast(tl.uint32, bitcast=True)
        else:
            rbits = tl.full(
                (BLOCK_SIZE_M, BLOCK_SIZE_N // 2),
                RBITS,
                dtype=tl.uint32,
            )

        x_e2m1 = tl.inline_asm_elementwise(
            asm="""
            {
            .reg .b16 tmp0, tmp1;
            cvt.rs.satfinite.e2m1x4.f32 tmp0, {$6, $2, $5, $1}, $9;
            cvt.rs.satfinite.e2m1x4.f32 tmp1, {$8, $4, $7, $3}, $10;
            mov.b32 $0, {tmp0, tmp1};
            }
            """,
            constraints="=r,r,r,r,r,r,r,r,r,r,r,r,r",
            args=[x_block_scaled_b1, x_block_scaled_b2, rbits],
            dtype=tl.uint8,
            is_pure=True,
            pack=4,
        )

    return x_e2m1, x_scales.reshape(4, 32, 4).permute(1, 0, 2).ravel()


@triton.jit
def nvfp4_fouroversix_quantization_kernel(  # noqa: C901, PLR0912, PLR0915
    x_block,
    x_amax_ptr,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    ROUND_STYLE: tl.constexpr,
    BLOCK_SCALE_2D: tl.constexpr,
    SCALE_RULE: tl.constexpr,
    RBITS: tl.constexpr,
) -> None:
    x_amax = tl.load(x_amax_ptr)
    x_scale_blocks = x_block.reshape(128, 4, 16)

    if x_amax == 0:
        x_scales_hp = tl.full((128, 4), 0, dtype=tl.float32)
    else:
        encode_scale = tl.div_rn(E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX, x_amax)
        x_scales_hp = (
            tl.div_rn(tl.max(x_scale_blocks.abs(), axis=-1), E2M1_MAX_VALUE)
            * encode_scale
        )

    if BLOCK_SCALE_2D:
        x_scales_hp = (
            tl.max(
                x_scales_hp.reshape(8, 16, 4).permute(0, 2, 1),
                axis=-1,
            )
            .expand_dims(0)
            .broadcast_to(16, 8, 4)
            .permute(1, 0, 2)
            .reshape(128, 4)
        )

    x_scales_6 = x_scales_hp.to(tl.float8e4nv)
    x_scales_4 = (x_scales_hp * 1.5).to(tl.float8e4nv)

    decode_scale = tl.div_rn(
        1,
        tl.div_rn(E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX, x_amax),
    )

    (x_block_scaled_6_b1, x_block_scaled_6_b2) = (
        tl.where(
            x_scales_6.expand_dims(2).to(x_amax.dtype) != 0,
            x_scale_blocks
            * tl.div_rn(1, decode_scale * x_scales_6.to(x_amax.dtype).expand_dims(2)),
            0,
        )
        .reshape(BLOCK_SIZE_M, BLOCK_SIZE_N // 2, 2)
        .split()
    )

    (x_block_scaled_4_b1, x_block_scaled_4_b2) = (
        tl.where(
            x_scales_4.expand_dims(2).to(x_amax.dtype) != 0,
            x_scale_blocks
            * tl.div_rn(1, decode_scale * x_scales_4.to(x_amax.dtype).expand_dims(2)),
            0,
        )
        .reshape(BLOCK_SIZE_M, BLOCK_SIZE_N // 2, 2)
        .split()
    )

    if ROUND_STYLE == ROUND_STYLE_NEAREST:
        (x_e2m1_6, x_e2m1_4, x_fp16x2_6, x_fp16x2_4) = tl.inline_asm_elementwise(
            asm="""
                {
                .reg .b8 byte0, byte1, byte2, byte3;

                cvt.rn.satfinite.e2m1x2.f32 byte0, $28, $20;
                cvt.rn.f16x2.e2m1x2 $4, byte0;
                cvt.rn.satfinite.e2m1x2.f32 byte1, $29, $21;
                cvt.rn.f16x2.e2m1x2 $5, byte1;
                cvt.rn.satfinite.e2m1x2.f32 byte2, $30, $22;
                cvt.rn.f16x2.e2m1x2 $6, byte2;
                cvt.rn.satfinite.e2m1x2.f32 byte3, $31, $23;
                cvt.rn.f16x2.e2m1x2 $7, byte3;
                mov.b32 $0, {byte0, byte1, byte2, byte3};

                cvt.rn.satfinite.e2m1x2.f32 byte0, $32, $24;
                cvt.rn.f16x2.e2m1x2 $8, byte0;
                cvt.rn.satfinite.e2m1x2.f32 byte1, $33, $25;
                cvt.rn.f16x2.e2m1x2 $9, byte1;
                cvt.rn.satfinite.e2m1x2.f32 byte2, $34, $26;
                cvt.rn.f16x2.e2m1x2 $10, byte2;
                cvt.rn.satfinite.e2m1x2.f32 byte3, $35, $27;
                cvt.rn.f16x2.e2m1x2 $11, byte3;
                mov.b32 $1, {byte0, byte1, byte2, byte3};

                cvt.rn.satfinite.e2m1x2.f32 byte0, $44, $36;
                cvt.rn.f16x2.e2m1x2 $12, byte0;
                cvt.rn.satfinite.e2m1x2.f32 byte1, $45, $37;
                cvt.rn.f16x2.e2m1x2 $13, byte1;
                cvt.rn.satfinite.e2m1x2.f32 byte2, $46, $38;
                cvt.rn.f16x2.e2m1x2 $14, byte2;
                cvt.rn.satfinite.e2m1x2.f32 byte3, $47, $39;
                cvt.rn.f16x2.e2m1x2 $15, byte3;
                mov.b32 $2, {byte0, byte1, byte2, byte3};

                cvt.rn.satfinite.e2m1x2.f32 byte0, $48, $40;
                cvt.rn.f16x2.e2m1x2 $16, byte0;
                cvt.rn.satfinite.e2m1x2.f32 byte1, $49, $41;
                cvt.rn.f16x2.e2m1x2 $17, byte1;
                cvt.rn.satfinite.e2m1x2.f32 byte2, $50, $42;
                cvt.rn.f16x2.e2m1x2 $18, byte2;
                cvt.rn.satfinite.e2m1x2.f32 byte3, $51, $43;
                cvt.rn.f16x2.e2m1x2 $19, byte3;
                mov.b32 $3, {byte0, byte1, byte2, byte3};
                }
                """,
            constraints="=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r",
            args=[
                x_block_scaled_6_b1,
                x_block_scaled_6_b2,
                x_block_scaled_4_b1,
                x_block_scaled_4_b2,
            ],
            dtype=(tl.uint8, tl.uint8, tl.uint32, tl.uint32),
            is_pure=True,
            pack=8,
        )
    elif ROUND_STYLE == ROUND_STYLE_STOCHASTIC:
        if RBITS == -1:
            rbits = tl.rand(
                0,
                tl.arange(0, BLOCK_SIZE_M)[:, None] * BLOCK_SIZE_N // 2
                + tl.arange(0, BLOCK_SIZE_N // 2)[None, :],
            ).cast(tl.uint32, bitcast=True)
        else:
            rbits = tl.full(
                (BLOCK_SIZE_M, BLOCK_SIZE_N // 2),
                RBITS,
                dtype=tl.uint32,
            )

        (x_e2m1_6, x_e2m1_4, x_fp16x2_6, x_fp16x2_4) = tl.inline_asm_elementwise(
            asm="""
                {
                .reg .b16 tmp0, tmp1;
                .reg .b8 byte0, byte1;

                cvt.rs.satfinite.e2m1x4.f32 tmp0, {$29, $21, $28, $20}, $52;
                mov.b16 {byte1, byte0}, tmp0;
                cvt.rn.f16x2.e2m1x2 $4, byte0;
                cvt.rn.f16x2.e2m1x2 $5, byte1;
                cvt.rs.satfinite.e2m1x4.f32 tmp1, {$31, $23, $30, $22}, $53;
                mov.b16 {byte1, byte0}, tmp1;
                cvt.rn.f16x2.e2m1x2 $6, byte0;
                cvt.rn.f16x2.e2m1x2 $7, byte1;
                mov.b32 $0, {tmp0, tmp1};

                cvt.rs.satfinite.e2m1x4.f32 tmp0, {$33, $25, $32, $24}, $54;
                mov.b16 {byte1, byte0}, tmp0;
                cvt.rn.f16x2.e2m1x2 $8, byte0;
                cvt.rn.f16x2.e2m1x2 $9, byte1;
                cvt.rs.satfinite.e2m1x4.f32 tmp1, {$35, $27, $34, $26}, $55;
                mov.b16 {byte1, byte0}, tmp1;
                cvt.rn.f16x2.e2m1x2 $10, byte0;
                cvt.rn.f16x2.e2m1x2 $11, byte1;
                mov.b32 $1, {tmp0, tmp1};

                cvt.rs.satfinite.e2m1x4.f32 tmp0, {$45, $37, $44, $36}, $56;
                mov.b16 {byte1, byte0}, tmp0;
                cvt.rn.f16x2.e2m1x2 $12, byte0;
                cvt.rn.f16x2.e2m1x2 $13, byte1;
                cvt.rs.satfinite.e2m1x4.f32 tmp1, {$47, $39, $46, $38}, $57;
                mov.b16 {byte1, byte0}, tmp1;
                cvt.rn.f16x2.e2m1x2 $14, byte0;
                cvt.rn.f16x2.e2m1x2 $15, byte1;
                mov.b32 $2, {tmp0, tmp1};

                cvt.rs.satfinite.e2m1x4.f32 tmp0, {$49, $41, $48, $40}, $58;
                mov.b16 {byte1, byte0}, tmp0;
                cvt.rn.f16x2.e2m1x2 $16, byte0;
                cvt.rn.f16x2.e2m1x2 $17, byte1;
                cvt.rs.satfinite.e2m1x4.f32 tmp1, {$51, $43, $50, $42}, $59;
                mov.b16 {byte1, byte0}, tmp1;
                cvt.rn.f16x2.e2m1x2 $18, byte0;
                cvt.rn.f16x2.e2m1x2 $19, byte1;
                mov.b32 $3, {tmp0, tmp1};
                }
                """,
            constraints="=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,=r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r,r",
            args=[
                x_block_scaled_6_b1,
                x_block_scaled_6_b2,
                x_block_scaled_4_b1,
                x_block_scaled_4_b2,
                rbits,
            ],
            dtype=(tl.uint8, tl.uint8, tl.uint32, tl.uint32),
            is_pure=True,
            pack=8,
        )

    x_fp16_6_lo = (
        (x_fp16x2_6 & 0xFFFF)
        .cast(tl.uint16)
        .cast(tl.float16, bitcast=True)
        .cast(x_amax.dtype)
    )
    x_fp16_6_hi = (
        (x_fp16x2_6 >> 16)
        .cast(tl.uint16)
        .cast(tl.float16, bitcast=True)
        .cast(x_amax.dtype)
    )
    x_hp_6 = tl.join(x_fp16_6_lo, x_fp16_6_hi).reshape(128, 4, 16)

    x_dequantized_6 = tl.div_rn(
        x_hp_6 * x_scales_6.to(x_amax.dtype).expand_dims(2) * x_amax,
        E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX,
    )

    x_fp16_4_lo = (
        (x_fp16x2_4 & 0xFFFF)
        .cast(tl.uint16)
        .cast(tl.float16, bitcast=True)
        .cast(x_amax.dtype)
    )
    x_fp16_4_hi = (
        (x_fp16x2_4 >> 16)
        .cast(tl.uint16)
        .cast(tl.float16, bitcast=True)
        .cast(x_amax.dtype)
    )
    x_hp_4 = tl.join(x_fp16_4_lo, x_fp16_4_hi).reshape(128, 4, 16)

    x_dequantized_4 = tl.div_rn(
        x_hp_4 * x_scales_4.to(x_amax.dtype).expand_dims(2) * x_amax,
        E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX,
    )

    diff_6 = x_dequantized_6 - x_scale_blocks
    diff_4 = x_dequantized_4 - x_scale_blocks

    if SCALE_RULE == SCALE_RULE_ABS_MAX:
        six_error = tl.max(tl.abs(diff_6), axis=-1)
        four_error = tl.max(tl.abs(diff_4), axis=-1)
    elif SCALE_RULE == SCALE_RULE_MAE:
        six_error = tl.sum(tl.abs(diff_6), axis=-1)
        four_error = tl.sum(tl.abs(diff_4), axis=-1)
    elif SCALE_RULE == SCALE_RULE_MSE:
        six_error = tl.sum(diff_6 * diff_6, axis=-1)
        four_error = tl.sum(diff_4 * diff_4, axis=-1)

    if BLOCK_SCALE_2D:
        six_error = six_error.reshape(8, 16, 4).permute(0, 2, 1)
        four_error = four_error.reshape(8, 16, 4).permute(0, 2, 1)

        if SCALE_RULE == SCALE_RULE_ABS_MAX:
            six_error = tl.max(six_error, axis=-1)
            four_error = tl.max(four_error, axis=-1)
        elif SCALE_RULE == SCALE_RULE_MAE or SCALE_RULE == SCALE_RULE_MSE:
            six_error = tl.sum(six_error, axis=-1)
            four_error = tl.sum(four_error, axis=-1)

        six_error = (
            six_error.expand_dims(0)
            .broadcast_to(16, 8, 4)
            .permute(1, 0, 2)
            .reshape(128, 4)
        )
        four_error = (
            four_error.expand_dims(0)
            .broadcast_to(16, 8, 4)
            .permute(1, 0, 2)
            .reshape(128, 4)
        )

    x_e2m1 = tl.where(
        (four_error < six_error).expand_dims(2),
        x_e2m1_4.reshape(128, 4, 8),
        x_e2m1_6.reshape(128, 4, 8),
    ).reshape(128, 32)

    x_scales = (
        tl.where(four_error < six_error, x_scales_4, x_scales_6)
        .reshape(4, 32, 4)
        .permute(1, 0, 2)
        .ravel()
    )

    return x_e2m1, x_scales


@triton.jit
def fp4_quantization_kernel(
    x_desc,
    x_amax_ptr,
    x_e2m1_desc,
    x_sf_desc,
    # Meta-parameters
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    TRANSPOSE: tl.constexpr,
    FP4_FORMAT: tl.constexpr,
    ROUND_STYLE: tl.constexpr,
    BLOCK_SCALE_2D: tl.constexpr,
    SCALE_RULE: tl.constexpr,
    RBITS: tl.constexpr,
) -> None:
    pid_m = tl.program_id(0)
    pid_n = tl.program_id(1)

    m_block_offset = pid_m * BLOCK_SIZE_M
    n_block_offset = pid_n * BLOCK_SIZE_N

    # Load [B, B] block from A or A^T
    if not TRANSPOSE:
        x_block = x_desc.load([m_block_offset, n_block_offset])
    else:
        x_block = x_desc.load([n_block_offset, m_block_offset]).T

    x_block = x_block.to(tl.float32)

    if SCALE_RULE == SCALE_RULE_STATIC_6 or SCALE_RULE == SCALE_RULE_STATIC_4:
        x_e2m1, x_scales = block_scaled_fp4_quantization_kernel(
            x_block,
            x_amax_ptr,
            BLOCK_SIZE_M,
            BLOCK_SIZE_N,
            FP4_FORMAT,
            ROUND_STYLE,
            BLOCK_SCALE_2D,
            SCALE_RULE,
            RBITS,
        )
    else:
        x_e2m1, x_scales = nvfp4_fouroversix_quantization_kernel(
            x_block,
            x_amax_ptr,
            BLOCK_SIZE_M,
            BLOCK_SIZE_N,
            ROUND_STYLE,
            BLOCK_SCALE_2D,
            SCALE_RULE,
            RBITS,
        )

    e2m1_n_block_offset = pid_n * BLOCK_SIZE_N // 2
    x_e2m1_desc.store([m_block_offset, e2m1_n_block_offset], x_e2m1)

    scale_block_offset = (pid_m * tl.num_programs(1) + pid_n) * SCALE_MEGABLOCK_SIZE
    x_sf_desc.store([scale_block_offset], x_scales)


def quantize_to_fp4(
    x: torch.Tensor,
    x_amax: torch.Tensor | None = None,
    had: torch.Tensor | None = None,
    *,
    fp4_format: DataType = DataType.nvfp4,
    round_style: RoundStyle = RoundStyle.nearest,
    scale_rule: ScaleRule = ScaleRule.mse,
    block_scale_2d: bool = False,
    transpose: bool = False,
    rbits: int = 0,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor | None]:
    if transpose:
        N, M = x.shape
    else:
        M, N = x.shape

    block_size_m = 128
    block_size_n = 4 * fp4_format.block_size()
    scale_dtype = torch.float8_e4m3fn if fp4_format == DataType.nvfp4 else torch.uint8

    if x_amax is None:
        x_amax = (
            x.abs().max().float()
            if fp4_format == DataType.nvfp4
            else torch.ones(1, device=x.device, dtype=torch.float32)
        )

    padded_m = M + (block_size_m - M % block_size_m) % block_size_m
    padded_n = N + (block_size_n - N % block_size_n) % block_size_n

    x_e2m1 = torch.empty((padded_m, padded_n // 2), device=x.device, dtype=torch.uint8)
    x_sf = torch.empty(
        padded_m * padded_n // fp4_format.block_size(),
        device=x.device,
        dtype=scale_dtype,
    )

    grid = lambda _: (  # noqa: E731
        padded_m // block_size_m,
        padded_n // block_size_n,
    )

    x_desc = TensorDescriptor.from_tensor(
        x,
        block_shape=[
            block_size_m if not transpose else block_size_n,
            block_size_n if not transpose else block_size_m,
        ],
    )
    x_e2m1_desc = TensorDescriptor.from_tensor(
        x_e2m1,
        block_shape=[block_size_m, block_size_n // 2],
    )
    x_sf_desc = TensorDescriptor.from_tensor(
        x_sf,
        block_shape=[SCALE_MEGABLOCK_SIZE.value],
    )

    if had is not None:
        had_block_size = had.shape[0]

        if M % had_block_size != 0:
            msg = (
                f"The first dimension of A ({M}) must be divisible by the width of H "
                f"({had_block_size})"
            )
            raise ValueError(msg)
        if N % had_block_size != 0:
            msg = (
                f"The second dimension of A ({N}) must be divisible by the width of H "
                f"({had_block_size})"
            )
            raise ValueError(msg)
        if had.shape[0] != had.shape[1]:
            msg = "H must be a square matrix"
            raise ValueError(msg)
        if (had.shape[0] & (had.shape[0] - 1)) != 0:
            msg = "H must have dimensions that are a power of two"
            raise ValueError(msg)

        x_rht = torch.empty((M, N), device=x.device, dtype=torch.bfloat16)

        h_desc = TensorDescriptor.from_tensor(
            had,
            block_shape=[had_block_size, had_block_size],
        )
        x_rht_desc = TensorDescriptor.from_tensor(
            x_rht,
            block_shape=[block_size_m, block_size_n],
        )

        rht_kernel[grid](
            x_desc,
            h_desc,
            x_rht_desc,
            BLOCK_SIZE_M=block_size_m,
            BLOCK_SIZE_N=block_size_n,
            TRANSPOSE=transpose,
        )

        transpose = False
        x_amax = x_rht.abs().max().float()

    fp4_quantization_kernel[grid](
        x_rht_desc if had is not None else x_desc,
        x_amax,
        x_e2m1_desc,
        x_sf_desc,
        BLOCK_SIZE_M=block_size_m,
        BLOCK_SIZE_N=block_size_n,
        TRANSPOSE=transpose,
        FP4_FORMAT=fp4_format.value,
        ROUND_STYLE=round_style.value,
        BLOCK_SCALE_2D=block_scale_2d,
        SCALE_RULE=scale_rule.value,
        RBITS=rbits,
    )

    if fp4_format == DataType.mxfp4:
        x_sf = x_sf.view(torch.float8_e8m0fnu)

    return x_e2m1, x_sf, x_amax



import torch
import torch.nn.functional as F
from fouroversix.quantize.backend import QuantizeBackendBase
from fouroversix.quantize.config import QuantizationConfig
from fouroversix.quantize.quantized_tensor import QuantizedTensor
from fouroversix.quantize.utils import get_rht_matrix

from .reference import quantize_to_fp4


class PyTorchQuantizeBackend(QuantizeBackendBase):
    """
    The PyTorch quantization backend. Supports all quantization options, and can be run
    on non-Blackwell GPUs, but is slow. Should be used primarily as a reference.
    """

    @classmethod
    def is_available(cls) -> bool:
        """Return True if the PyTorch backend is available on the current machine."""
        return True

    @classmethod
    def is_supported(
        cls,
        x: torch.Tensor,  # noqa: ARG003
        config: QuantizationConfig,  # noqa: ARG003
    ) -> bool:
        """
        Return True if the PyTorch backend supports the given input and quantization
        configuration.
        """

        return True

    @classmethod
    def quantize_to_fp4(
        cls,
        x: torch.Tensor,
        config: QuantizationConfig,
    ) -> QuantizedTensor:
        """
        Quantize a tensor to FP4 using the PyTorch backend.

        Args:
            x (torch.Tensor): The input tensor to quantize.
            config (QuantizationConfig): The quantization configuration.

        Returns:
            The quantized tensor.

        """

        input_shape = (x.shape[1], x.shape[0]) if config.transpose else x.shape

        rows_div = 128
        cols_div = 4 * config.dtype.block_size()

        if input_shape[0] % rows_div != 0 or input_shape[1] % cols_div != 0:
            x = F.pad(
                x,
                (
                    0,
                    (
                        cols_div - (input_shape[1] % cols_div)
                        if input_shape[1] % cols_div > 0
                        else 0
                    ),
                    0,
                    (
                        rows_div - (input_shape[0] % rows_div)
                        if input_shape[0] % rows_div > 0
                        else 0
                    ),
                ),
            )

        if x.device.type == "meta":
            values = torch.zeros(
                input_shape[0],
                input_shape[1] // 2,
                device=x.device,
                dtype=torch.uint8,
            )
            scale_factors = torch.zeros(
                input_shape[0] * input_shape[1] // config.dtype.block_size(),
                device=x.device,
                dtype=(
                    torch.uint8
                    if config.dtype.scale_dtype() == torch.float8_e8m0fnu
                    else config.dtype.scale_dtype()
                ),
            )
            amax = torch.zeros(1, device=x.device, dtype=torch.float32)
        else:
            values, scale_factors, amax = quantize_to_fp4(
                x,
                had=get_rht_matrix() if config.rht else None,
                fp4_format=config.dtype,
                round_style=config.round_style,
                scale_rule=config.scale_rule,
                block_scale_2d=config.block_scale_2d,
                transpose=config.transpose,
            )

        return QuantizedTensor(
            values,
            scale_factors,
            amax,
            config.dtype,
            input_shape,
            config.scale_rule,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

import torch.nn as nn

if TYPE_CHECKING:
    from collections.abc import Callable

    from .config import ModelQuantizationConfig


class QuantizedModule:
    """Base class for all quantized modules."""

    _registry: ClassVar[dict[type[nn.Module], type[nn.Module]]] = {}
    _should_replace_existing_modules_in_model: ClassVar[dict[type[nn.Module], bool]] = (
        {}
    )

    @classmethod
    def is_quantized_module_type(cls, module_type: type[nn.Module]) -> bool:
        """Return True if the given module type is a quantized module."""
        return module_type in cls._registry.values()

    @classmethod
    def get_cls(
        cls,
        high_precision_cls: type[nn.Module],
    ) -> type[nn.Module] | None:
        """Get the quantized module for a given high-precision module."""
        return cls._registry.get(high_precision_cls)

    @classmethod
    def should_replace_existing_modules_in_model(
        cls,
        module_type: type[nn.Module],
    ) -> bool:
        """Determine whether module should be replaced."""
        return cls._should_replace_existing_modules_in_model.get(module_type, False)

    @classmethod
    def register(
        cls,
        high_precision_cls: type[nn.Module],
        *,
        replace_existing_modules_in_registry: bool = False,
        replace_existing_modules_in_model: bool = True,
    ) -> Callable[[type[nn.Module]], type[nn.Module]]:
        """
        Register a new type of quantized module.

        Args:
            high_precision_cls: (`type[nn.Module]`): The high precision module to be
            mapped to a fouroversix quantized module.
            replace_existing_modules_in_registry (bool): determines whether we should
            replace the existing module in the registry.
            replace_existing_modules_in_model (bool): determines whether we should
            replace the existing module in the model including the weights.

        """

        if (
            high_precision_cls in cls._registry
            and not replace_existing_modules_in_registry
        ):
            msg = f"High-precision module {high_precision_cls} is already registered."
            raise ValueError(msg)

        modules_to_delete = []

        for module_cls in cls._registry:
            if high_precision_cls is not None and issubclass(
                high_precision_cls,
                module_cls,
            ):
                if replace_existing_modules_in_registry:
                    modules_to_delete.append(module_cls)
                else:
                    msg = (
                        f"High-precision module {high_precision_cls} is a subclass of "
                        f"{module_cls}, which is already registered."
                    )
                    raise TypeError(msg)

        for module_cls in modules_to_delete:
            del cls._registry[module_cls]

        def inner_wrapper(
            wrapped_cls: type[nn.Module],
        ) -> type[nn.Module]:
            cls._registry[high_precision_cls] = wrapped_cls
            cls._should_replace_existing_modules_in_model[high_precision_cls] = (
                replace_existing_modules_in_model
            )
            return wrapped_cls

        return inner_wrapper


def quantize_model(
    model: nn.Module,
    config: ModelQuantizationConfig,
    **kwargs: dict[str, Any],
) -> None:
    for module_name, module in model.named_modules():
        if (
            module_name == ""
            or module_name in config.modules_to_not_convert
            or not isinstance(module, nn.Module)
        ):
            continue

        module_cls = QuantizedModule.get_cls(type(module))
        should_replace = QuantizedModule.should_replace_existing_modules_in_model(
            type(module),
        )

        if module_cls is None or not should_replace:
            continue

        quantized_module = module_cls(
            module,
            config.get_module_config(module_name),
            **kwargs,
        )
        model.set_submodule(module_name, quantized_module)



from .gpt_oss import FourOverSixGptOssMLP
from .linear import FourOverSixLinear
from .qwen import FourOverSixQwenExperts

__all__ = ["FourOverSixGptOssMLP", "FourOverSixLinear", "FourOverSixQwenExperts"]



from collections.abc import Callable
from typing import ClassVar

from transformers import WeightConverter


class WeightConversions:
    """Base class for weight conversions for quantized models."""

    _registry: ClassVar[dict[str, list[WeightConverter]]] = {}

    @classmethod
    def register(
        cls,
        pre_quantized_model_config_type: str,
    ) -> Callable[[type], list[WeightConverter]]:
        """Register a new type of weight conversion."""
        if pre_quantized_model_config_type in cls._registry:
            msg = f"Model with config {pre_quantized_model_config_type} is \
            already registered."
            raise ValueError(msg)

        def inner_wrapper(
            wrapped_cls: type,
        ) -> list[WeightConverter]:
            weight_conversions = wrapped_cls.get_weight_conversions()
            cls._registry[pre_quantized_model_config_type] = weight_conversions
            return weight_conversions

        return inner_wrapper

    @classmethod
    def get_weight_conversions(
        cls,
        pre_quantized_model_config_type: str,
    ) -> list[WeightConverter]:
        """
        Get the weight conversion for a given model type determined
        by the model config type.
        """
        weight_conversions = cls._registry.get(pre_quantized_model_config_type, None)
        if weight_conversions is None:
            msg = f"Config type {pre_quantized_model_config_type} not supported."
            raise ValueError(msg)
        return weight_conversions



from .config import QuantizationConfig
from .frontend import quantize_to_fp4
from .quantized_tensor import QuantizedTensor
from .utils import get_rht_matrix

__all__ = ["QuantizationConfig", "QuantizedTensor", "get_rht_matrix", "quantize_to_fp4"]



from .config import ModelQuantizationConfig, ModuleQuantizationConfig
from .modules import FourOverSixLinear
from .quantize import QuantizedModule, quantize_model

__all__ = [
    "FourOverSixLinear",
    "ModelQuantizationConfig",
    "ModuleQuantizationConfig",
    "QuantizedModule",
    "quantize_model",
]



from abc import ABC, abstractmethod

import torch
from fouroversix.utils import DataType, ScaleRule

from .config import QuantizationConfig
from .quantized_tensor import QuantizedTensor


class QuantizeBackendBase(ABC):
    """Base class for all quantization backends."""

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """Return True if the backend is available on the current machine."""
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)

    @classmethod
    @abstractmethod
    def is_supported(cls, x: torch.Tensor, config: QuantizationConfig) -> bool:
        """
        Return True if the backend supports the given input and quantization
        configuration.
        """

        if not cls.is_available():
            return False

        if x.ndim != 2:  # noqa: PLR2004
            return False

        if config.dtype not in {DataType.mxfp4, DataType.nvfp4}:
            return False

        if config.dtype == DataType.mxfp4 and config.scale_rule not in {
            ScaleRule.static_6,
            ScaleRule.static_4,
        }:
            msg = (
                "MXFP4 quantization only supports the `static_6` and `static_4` scale "
                "rules"
            )
            raise ValueError(msg)

        return True

    @classmethod
    @abstractmethod
    def quantize_to_fp4(
        cls,
        x: torch.Tensor,
        config: QuantizationConfig,
    ) -> QuantizedTensor:
        """
        Quantize a tensor to FP4 using the backend.

        Args:
            x (torch.Tensor): The input tensor to quantize.
            config (QuantizationConfig): The quantization configuration.

        Returns:
            The quantized tensor.

        """

        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)



import torch

import fouroversix._C  # noqa: F401


def gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.bfloat16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake("fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt")
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.float16, device=a.device)


def gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,
    b_sf: torch.Tensor,
    alpha: torch.Tensor,
) -> torch.Tensor:
    return torch.ops.fouroversix.gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120.default(
        a,
        b,
        a_sf,
        b_sf,
        alpha,
    )


@torch.library.register_fake(
    "fouroversix::gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120",
)
def _(
    a: torch.Tensor,
    b: torch.Tensor,
    a_sf: torch.Tensor,  # noqa: ARG001
    b_sf: torch.Tensor,  # noqa: ARG001
    alpha: torch.Tensor,  # noqa: ARG001
) -> torch.Tensor:
    m = a.shape[0]
    n = b.shape[0]
    return torch.empty(m, n, dtype=torch.float16, device=a.device)



from dataclasses import dataclass

from fouroversix.utils import DataType, QuantizeBackend, RoundStyle, ScaleRule


@dataclass
class QuantizationConfig:
    """
    Configuration to use when quantizing a tensor.

    Args:
        backend (QuantizeBackend): The backend to use for quantization. If no backend is
            provided, a backend will be selected automatically based on the available
            GPU and the specified options.
        block_scale_2d (bool): If True, scale factors will be computed across 16x16
            chunks of the input rather than 1x16 chunks. This is useful to apply to the
            weight matrix during training, so that W and W.T will be equivalent after
            quantization.
        dtype (DataType): The data type to quantize to.
        rbits (int): Random bits to provide to the cvt.rs instruction when performing
            stochastic rounding. Only supported with the Triton and CUDA backends. If
            set to -1, random bits will be generated on the fly.
        rht (bool): If True, the random Hadamard transform will be applied to the input
            prior to quantization.
        round_style (RoundStyle): The rounding style to apply during quantization.
        scale_rule (ScaleRule): The scaling rule to use during quantization.
        transpose (bool): If True, the output will be a quantized version of the
            transposed input. This may be helpful for certain operations during training
            as `fp4_matmul` requires that both tensors are provided in row-major format.

    """

    backend: QuantizeBackend | None = None
    block_scale_2d: bool = False
    dtype: DataType = DataType.nvfp4
    rbits: int = -1
    rht: bool = False
    round_style: RoundStyle = RoundStyle.nearest
    scale_rule: ScaleRule = ScaleRule.mse
    transpose: bool = False

    def __post_init__(self) -> None:
        """Convert string values to enums."""

        if isinstance(self.backend, str):
            self.backend = QuantizeBackend(self.backend)

        if isinstance(self.dtype, str):
            self.dtype = DataType(self.dtype)

        if isinstance(self.round_style, str):
            self.round_style = RoundStyle(self.round_style)

        if isinstance(self.scale_rule, str):
            self.scale_rule = ScaleRule(self.scale_rule)



from abc import ABC, abstractmethod

import torch
from fouroversix.quantize import QuantizedTensor
from fouroversix.utils import DataType


class MatmulBackendBase(ABC):
    """Base class for all matrix multiplication backends."""

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        """Return True if the backend is available on the current machine."""
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)

    @classmethod
    @abstractmethod
    def is_supported(
        cls,
        input: QuantizedTensor,
        other: QuantizedTensor,
        *,
        out_dtype: DataType,
    ) -> bool:
        """Return True if the backend supports the given inputs and output data type."""

        if not cls.is_available():
            return False

        if input.dtype != other.dtype:
            msg = "Both inputs must have the same dtype"
            raise ValueError(msg)

        if input.original_shape[1] != other.original_shape[1]:
            msg = (
                "The first input must be in row-major layout, the second input must be"
                "in column-major layout, and both inputs must have the same inner "
                "dimension"
            )
            raise ValueError(msg)

        return True

    @classmethod
    @abstractmethod
    def fp4_matmul(
        cls,
        input: QuantizedTensor,
        other: QuantizedTensor,
        *,
        out_dtype: DataType,
    ) -> torch.Tensor:
        """
        Perform a matrix multiplication (`a @ b.T`) between two quantized tensors using
        the backend.
        """
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)




import torch
from fouroversix.quantize import QuantizedTensor
from fouroversix.utils import DataType

from .backend import MatmulBackendBase


class PyTorchMatmulBackend(MatmulBackendBase):
    """
    The PyTorch matrix multiplication backend. Dequantizes both inputs to FP32 and
    performs an FP32 matrix multiplication in order to simulate an NVFP4 matrix
    multiplication which accumulates in FP32. Slow, but can be run on any GPU.
    """

    @classmethod
    def is_available(cls) -> bool:
        """Return True if the PyTorch backend is available on the current machine."""
        return True

    @classmethod
    def fp4_matmul(
        cls,
        input: QuantizedTensor,
        other: QuantizedTensor,
        *,
        out_dtype: DataType,
    ) -> torch.Tensor:
        """Perform a matrix multiplication (`a @ b.T`) between two quantized tensors."""

        out_shape = (input.original_shape[0], other.original_shape[0])

        out = torch.matmul(
            input.dequantize(dtype=torch.float32),
            other.dequantize(dtype=torch.float32).T,
        ).to(out_dtype.torch_dtype())

        if out.shape != out_shape:
            out = out[: out_shape[0], : out_shape[1]]

        return out



import functools
import math

import torch

"""
Credit: TransformerEngine
https://github.com/NVIDIA/TransformerEngine/blob/main/transformer_engine/pytorch/tensor/nvfp4_tensor.py
"""

HADAMARD_DIMENSION = 16


def get_no_random_sign_vector(device: int) -> torch.Tensor:
    """Non-random sign vector for Hadamard transform."""
    return torch.tensor([1], dtype=torch.float32, device=device)


def get_wgrad_sign_vector(device: int) -> torch.Tensor:
    """
    Hard-coded random signs for Hadamard transform.

    https://xkcd.com/221/

    """
    return torch.tensor(
        [1, 1, 1, -1, 1, -1, -1, -1, -1, -1, -1, 1, -1, 1, -1, -1],
        dtype=torch.float32,
        device=device,
    )


def get_hadamard_matrix(hadamard_dimension: int, device: int) -> torch.Tensor:
    """Construct a 16x16 Hadamard matrix."""

    if hadamard_dimension != HADAMARD_DIMENSION:
        msg = f"Only hadamard dimension {HADAMARD_DIMENSION} is supported."
        raise ValueError(msg)

    hadamard_scale = 1 / math.sqrt(hadamard_dimension)
    return (
        torch.tensor(
            [
                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                [1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1, 1, -1],
                [1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1],
                [1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1, 1, -1, -1, 1],
                [1, 1, 1, 1, -1, -1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1],
                [1, -1, 1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, 1, -1, 1],
                [1, 1, -1, -1, -1, -1, 1, 1, 1, 1, -1, -1, -1, -1, 1, 1],
                [1, -1, -1, 1, -1, 1, 1, -1, 1, -1, -1, 1, -1, 1, 1, -1],
                [1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1],
                [1, -1, 1, -1, 1, -1, 1, -1, -1, 1, -1, 1, -1, 1, -1, 1],
                [1, 1, -1, -1, 1, 1, -1, -1, -1, -1, 1, 1, -1, -1, 1, 1],
                [1, -1, -1, 1, 1, -1, -1, 1, -1, 1, 1, -1, -1, 1, 1, -1],
                [1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1],
                [1, -1, 1, -1, -1, 1, -1, 1, -1, 1, -1, 1, 1, -1, 1, -1],
                [1, 1, -1, -1, -1, -1, 1, 1, -1, -1, 1, 1, 1, 1, -1, -1],
                [1, -1, -1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 1, -1, -1, 1],
            ],
            dtype=torch.float32,
            device=device,
        )
        * hadamard_scale
    )


@functools.cache
def get_rht_matrix(
    *,
    with_random_sign_mask: bool = True,
    device: str | int = "cuda",
) -> torch.Tensor:
    """Construct matrix used in random Hadamard transform."""
    if with_random_sign_mask:
        signs = get_wgrad_sign_vector(device=device)
    else:
        signs = get_no_random_sign_vector(device=device)
    sign_matrix = signs * torch.eye(
        HADAMARD_DIMENSION,
        dtype=torch.float32,
        device=device,
    )
    rht_matrix = sign_matrix @ get_hadamard_matrix(HADAMARD_DIMENSION, device=device)
    return rht_matrix.to(dtype=torch.bfloat16)


def to_blocked(a: torch.Tensor) -> torch.Tensor:
    return (
        a.view(a.shape[0] // 128, 128, a.shape[1] // 4, 4)
        .transpose(1, 2)
        .reshape(-1, 4, 32, 4)
        .transpose(1, 2)
        .reshape(-1, 32, 16)
        .flatten()
    )



from dataclasses import dataclass

import torch
import torch.nn.functional as F
from fouroversix.utils import DataType, ScaleRule

from .utils import to_blocked


def from_blocked(a: torch.Tensor, orig_shape: tuple[int, int]) -> torch.Tensor:
    rows, cols = orig_shape
    return (
        a.view(-1, 32, 4, 4)
        .transpose(1, 2)
        .reshape(-1, cols // 4, 128, 4)
        .transpose(1, 2)
        .reshape(rows, cols)
    )


def convert_e2m1_to_fp8_e4m3(x: torch.Tensor) -> torch.Tensor:
    sign = (x >> 3) & 0x1
    exponent = (x >> 1) & 0x3
    mantissa = x & 0x1

    # Make adjustments
    new_exponent = torch.where(
        (exponent == 0) & (mantissa == 0),
        0,
        (exponent + 6) & 0xF,
    )
    new_mantissa = torch.where(exponent == 0, 0, mantissa << 2)

    return ((sign << 7) | (new_exponent << 3) | new_mantissa).view(torch.float8_e4m3fn)


def unpack_packed_fp4(
    x: torch.Tensor,
    to_dtype: torch.dtype = torch.float8_e4m3fn,
) -> torch.Tensor:
    if to_dtype == torch.float8_e4m3fn:
        convert_function = convert_e2m1_to_fp8_e4m3
    else:
        msg = f"Unsupported dtype: {to_dtype}"
        raise ValueError(msg)

    high = (x >> 4) & 0xF
    low = x & 0xF

    return torch.stack(
        [convert_function(low), convert_function(high)],
        dim=-1,
    ).reshape(x.shape[0], x.shape[1] * 2)


@dataclass
class QuantizedTensor:
    """A quantized tensor."""

    values: torch.Tensor
    scale_factors: torch.Tensor
    amax: torch.Tensor

    dtype: DataType
    original_shape: tuple[int, int]
    scale_rule: ScaleRule

    padded_shape: tuple[int, int]

    def __init__(
        self,
        values: torch.Tensor,
        scale_factors: torch.Tensor,
        amax: torch.Tensor,
        dtype: DataType,
        original_shape: tuple[int, int],
        scale_rule: ScaleRule,
        padded_shape: tuple[int, int] | None = None,
    ) -> None:
        super().__init__()

        if isinstance(dtype, str):
            dtype = DataType(dtype)

        if isinstance(original_shape, torch.Size):
            original_shape = tuple(original_shape)

        if isinstance(scale_rule, str):
            scale_rule = ScaleRule(scale_rule)

        if isinstance(padded_shape, torch.Size):
            padded_shape = tuple(padded_shape)

        self.dtype = dtype
        self.original_shape = original_shape
        self.scale_rule = scale_rule
        self.padded_shape = padded_shape

        if self.padded_shape is None:
            rows_div = 128

            # The scale factor layout requires 4 blocks along the K dimension for both
            # MXFP4 and NVFP4. See:
            # https://docs.nvidia.com/cutlass/latest/media/docs/cpp/blackwell_functionality.html#scale-factor-layouts
            cols_div = 4 * dtype.block_size()

            self.padded_shape = (
                original_shape[0]
                + (rows_div - original_shape[0] % rows_div) % rows_div,
                original_shape[1]
                + (cols_div - original_shape[1] % cols_div) % cols_div,
            )

            expected_packed_elements = self.padded_shape[0] * self.padded_shape[1] // 2
            expected_scale_factors = expected_packed_elements * 2 // dtype.block_size()

            if values.numel() != expected_packed_elements:
                values = F.pad(
                    values,
                    (
                        0,
                        # Divide by 2 because these are packed values
                        self.padded_shape[1] // 2 - values.shape[1],
                        0,
                        self.padded_shape[0] - values.shape[0],
                    ),
                )

            # If the scale factors are 1D, we assume that they are already in the
            # correct layout for Blackwell. See:
            # https://docs.nvidia.com/cutlass/latest/media/docs/cpp/blackwell_functionality.html#scale-factor-layouts
            if (
                scale_factors.ndim > 1
                and scale_factors.numel() != expected_scale_factors
            ):
                scale_factors = F.pad(
                    scale_factors,
                    (
                        0,
                        (
                            self.padded_shape[1] // dtype.block_size()
                            - scale_factors.shape[1]
                        ),
                        0,
                        self.padded_shape[0] - scale_factors.shape[0],
                    ),
                    value=0 if dtype == DataType.nvfp4 else 1,
                )

                scale_factors = to_blocked(scale_factors)

            if values.numel() != expected_packed_elements:
                msg = (
                    f"Expected {expected_packed_elements} e2m1 values, got "
                    f"{values.numel()}"
                )
                raise ValueError(msg)

            if scale_factors.numel() != expected_scale_factors:
                msg = (
                    f"Expected {expected_scale_factors} scale factors, got "
                    f"{scale_factors.numel()}"
                )
                raise ValueError(msg)

        self.values = values
        self.scale_factors = scale_factors
        self.amax = amax

    def dequantize(self, dtype: torch.dtype = torch.bfloat16) -> torch.Tensor:
        """Return a high-precision tensor with the dequantized values."""

        values = unpack_packed_fp4(self.values).to(dtype)
        scales = from_blocked(
            self.scale_factors,
            (
                self.padded_shape[0],
                self.padded_shape[1] // self.dtype.block_size(),
            ),
        )

        result = values * scales.to(dtype).repeat_interleave(
            self.dtype.block_size(),
            -1,
        )

        if self.dtype == DataType.nvfp4 and self.amax is not None:
            result = (
                result.to(torch.float32)
                * self.amax
                / (
                    self.scale_rule.max_allowed_e2m1_value()
                    * self.scale_rule.max_allowed_e4m3_value()
                )
            ).to(dtype)

        if result.shape != self.original_shape:
            result = result[: self.original_shape[0], : self.original_shape[1]]

        return result

    @property
    def device(self) -> torch.device:
        """Get device of the values in this tensor."""
        return self.values.device



from importlib.metadata import version

from .matmul import fp4_matmul
from .model import (
    FourOverSixLinear,
    ModelQuantizationConfig,
    ModuleQuantizationConfig,
    QuantizedModule,
    quantize_model,
)
from .quantize import QuantizationConfig, QuantizedTensor, quantize_to_fp4
from .utils import DataType, MatmulBackend, QuantizeBackend, RoundStyle, ScaleRule
from .weight_conversions import WeightConversions

__version__ = version("fouroversix")

__all__ = [
    "DataType",
    "FourOverSixLinear",
    "MatmulBackend",
    "ModelQuantizationConfig",
    "ModuleQuantizationConfig",
    "QuantizationConfig",
    "QuantizeBackend",
    "QuantizedModule",
    "QuantizedTensor",
    "RoundStyle",
    "ScaleRule",
    "WeightConversions",
    "fp4_matmul",
    "quantize_model",
    "quantize_to_fp4",
]



import torch
from fouroversix.utils import QuantizeBackend

from .config import QuantizationConfig
from .cuda import CUDAQuantizeBackend
from .pytorch import PyTorchQuantizeBackend
from .quantized_tensor import QuantizedTensor
from .transformer_engine import TransformerEngineQuantizeBackend
from .triton import TritonQuantizeBackend

AVAILABLE_BACKENDS = {
    QuantizeBackend.cuda: CUDAQuantizeBackend,
    QuantizeBackend.transformer_engine: TransformerEngineQuantizeBackend,
    QuantizeBackend.triton: TritonQuantizeBackend,
    QuantizeBackend.pytorch: PyTorchQuantizeBackend,
}


def quantize_to_fp4(
    x: torch.Tensor,
    config: QuantizationConfig | None = None,
) -> QuantizedTensor:
    """
    Quantize a tensor to FP4.

    ## Sample Code

    ### With Four Over Six

    ```python
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    x_quantized = quantize_to_fp4(x)
    ```

    ### Without Four Over Six

    ```python
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    config = QuantizationConfig(scale_rule="static_6")
    x_quantized = quantize_to_fp4(x, config)
    ```

    ### With Stochastic Rounding

    ```python
    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    config = QuantizationConfig(round_style="stochastic")
    x_quantized = quantize_to_fp4(x, config)
    ```

    ### With the Random Hadamard Transform

    ```python
    from fouroversix.quantize import get_rht_matrix

    x = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    config = QuantizationConfig(rht=True)
    x_quantized = quantize_to_fp4(x, config)
    ```

    ## Backends

    We provide three different implementations of FP4 quantization:

    - **CUDA**: A fast implementation written in CUDA which currently only supports
        basic quantization options (no 2D block scaling, no stochastic rounding, no
        random Hadamard transform). Can be used for inference, but not training.
        Requires a Blackwell GPU.
    - **Triton**: A slightly slower implementation written in Triton which supports all
        operations needed for training. Also requires a Blackwell GPU.
    - **PyTorch**: A slow implementation written in PyTorch which supports all
        operations and can be run on any GPU.

    If `quantize_to_fp4` is called with `backend=None`, a backend will be selected
    automatically based on the following rules:

    - If there is no GPU available, or if the available GPU is not a Blackwell GPU,
        select PyTorch.
    - If any quantization options are set other than `scale_rule`, select Triton.
        - However, if the available GPU is SM120 (i.e. RTX 5090, RTX 6000) and
            `round_style` is set to `RoundStyle.stochastic`, select PyTorch as
            stochastic rounding does not have hardware support on SM120 GPUs.
    - Otherwise, select CUDA.

    ## Parameters

    Args:
        x (torch.Tensor): The input tensor to quantize.
        config (QuantizationConfig): The quantization configuration to use. If no
            configuration is provided, a default configuration will be used (NVFP4,
            1D block scaling, round-to-nearest, and 4/6 with the MSE selection rule).

    Returns:
        The quantized tensor.

    """

    if config is None:
        config = QuantizationConfig()

    selected_backend = config.backend

    if selected_backend is None:
        for backend in [
            QuantizeBackend.cuda,
            QuantizeBackend.triton,
            QuantizeBackend.pytorch,
        ]:
            if AVAILABLE_BACKENDS[backend] is not None and AVAILABLE_BACKENDS[
                backend
            ].is_supported(x, config):
                selected_backend = backend
                break
        else:
            msg = "No backend found that supports the given parameters"
            raise ValueError(msg)

    elif not AVAILABLE_BACKENDS[selected_backend].is_supported(x, config):
        msg = f"Backend {selected_backend} does not support the given parameters"
        raise ValueError(msg)

    return AVAILABLE_BACKENDS[selected_backend].quantize_to_fp4(x, config)



from enum import Enum

import torch

SM_100 = 10
SM_110 = 11
SM_120 = 12

BLACKWELL_SM_IDS = {SM_100, SM_110, SM_120}


class DataType(str, Enum):
    """Data types."""

    bfloat16 = "bfloat16"
    float16 = "float16"
    float32 = "float32"
    mxfp4 = "mxfp4"
    nvfp4 = "nvfp4"

    def block_size(self) -> int | None:
        """Return the block size if this a block-scaled format, or `None` otherwise."""

        return {
            DataType.mxfp4: 32,
            DataType.nvfp4: 16,
        }.get(self)

    def scale_dtype(self) -> torch.dtype | None:
        """Return the scale dtype if this a block-scaled format, or `None` otherwise."""

        return {
            DataType.mxfp4: torch.float8_e8m0fnu,
            DataType.nvfp4: torch.float8_e4m3fn,
        }.get(self)

    def torch_dtype(self) -> torch.dtype | None:
        """
        Return the corresponding torch.dtype if one is available, or `None`
        otherwise.
        """

        return {
            DataType.bfloat16: torch.bfloat16,
            DataType.float16: torch.float16,
            DataType.float32: torch.float32,
        }.get(self)


class MatmulBackend(str, Enum):
    """
    Backends for matrix multiplication with FP4.

    - `cutlass`: CUTLASS implementation. This requires a Blackwell GPU.
    - `pytorch`: PyTorch implementation which first dequantizes the input tensors to
        FP32 and then performs an FP32 matrix multiplication.
    """

    cutlass = "cutlass"
    pytorch = "pytorch"


class QuantizeBackend(str, Enum):
    """
    Backends for quantizing a tensor to NVFP4 or MXFP4.

    - `cuda`: CUDA implementation. Requires a Blackwell GPU, and currently only supports
        the forward pass for PTQ (no stochastic rounding, no transposed matrices, no
        RHT, no 2D block scaling).
    - `pytorch`: PyTorch implementation.
    - `triton`: Triton implementation. Requires a Blackwell GPU.
    """

    cuda = "cuda"
    pytorch = "pytorch"
    transformer_engine = "transformer_engine"
    triton = "triton"


class RoundStyle(str, Enum):
    """
    Rounding styles for quantization.

    - `nearest`: Round to the nearest FP4 value.
    - `stochastic`: Round to the nearest FP4 value after applying random noise to each
        value.
    """

    nearest = "nearest"
    stochastic = "stochastic"


class ScaleRule(str, Enum):
    """
    Block scale selection rules for NVFP4 quantization.

    - `abs_max`: Between 4 and 6, select the block scale that minimizes the maximum
        absolute quantization error.
    - `static_4`: Select 4 for all blocks.
    - `static_6`: Select 6 for all blocks (normal NVFP4 quantization).
    - `mae`: Between 4 and 6, select the block scale that minimizes the mean absolute
        quantization error.
    - `mse`: Between 4 and 6, select the block scale that minimizes the mean squared
        quantization error.
    """

    abs_max = "abs_max"
    mae = "mae"
    mse = "mse"
    static_4 = "static_4"
    static_6 = "static_6"

    def cuda_id(self) -> int:
        """ID for the rule in the CUDA implementation."""

        return {
            ScaleRule.abs_max: 4,
            ScaleRule.mae: 2,
            ScaleRule.mse: 3,
            ScaleRule.static_4: 1,
            ScaleRule.static_6: 0,
        }[self]

    def is_static(self) -> bool:
        """Return True if the rule is static, False otherwise."""
        return self in {ScaleRule.static_4, ScaleRule.static_6}

    def max_allowed_e2m1_value(self) -> int:
        """Return the maximum allowed E2M1 value for the rule."""
        return 4 if self == ScaleRule.static_4 else 6

    def max_allowed_e4m3_value(self) -> int:
        """Return the maximum allowed E4M3 value for the rule."""
        return 448 if self in {ScaleRule.static_6, ScaleRule.static_4} else 256



from dataclasses import dataclass, field
from typing import Any

from fouroversix.quantize import QuantizationConfig
from fouroversix.utils import DataType, MatmulBackend, QuantizeBackend, ScaleRule


@dataclass
class ModuleQuantizationConfig:
    """
    Configuration for quantizing modules with Four Over Six.

    Args:
        activation_scale_rule (ScaleRule | None): The scaling rule to use for activation
            tensors. If not provided, `scale_rule` will be used.
        dtype (DataType): The quantization data type to use for the module. Defaults to
            `DataType.nvfp4`.
        gradient_scale_rule (ScaleRule | None): The scaling rule to use for gradient
            tensors. If not provided, `scale_rule` will be used.
        keep_master_weights (bool): Whether to keep the master weights. Defaults to
            `False`.
        matmul_backend (MatmulBackend | None): The backend to use for matrix
            multiplications. If not provided, a backend will be selected automatically
            based on the available GPU and the specified options.
        output_dtype (DataType): The data type to use for the module's output. Defaults
            to `DataType.bfloat16`.
        quantize_backend (QuantizeBackend | None): The backend to use for quantization.
            If not provided, a backend will be selected automatically based on the
            available GPU and the specified options.
        scale_rule (ScaleRule): The fallback scaling rule which will be used if any of
            the other scaling rules are not specified.
        weight_scale_2d (bool): Whether to use 2D block scaling for weights. Should be
            set to `True` if the module is used for training.
        weight_scale_rule (ScaleRule | None): The scaling rule to use for weights. If
            not provided, `scale_rule` will be used.

    """

    activation_scale_rule: ScaleRule | None = None
    dtype: DataType = DataType.nvfp4
    gradient_scale_rule: ScaleRule | None = None
    keep_master_weights: bool = False
    matmul_backend: MatmulBackend | None = None
    output_dtype: DataType = DataType.bfloat16
    quantize_backend: QuantizeBackend | None = None
    scale_rule: ScaleRule = ScaleRule.mse
    weight_scale_2d: bool = False
    weight_scale_rule: ScaleRule | None = None

    def __post_init__(self) -> None:
        """Convert string values to enums."""

        if isinstance(self.activation_scale_rule, str):
            self.activation_scale_rule = ScaleRule(self.activation_scale_rule)

        if isinstance(self.dtype, str):
            self.dtype = DataType(self.dtype)

        if isinstance(self.gradient_scale_rule, str):
            self.gradient_scale_rule = ScaleRule(self.gradient_scale_rule)

        if isinstance(self.matmul_backend, str):
            self.matmul_backend = MatmulBackend(self.matmul_backend)

        if isinstance(self.output_dtype, str):
            self.output_dtype = DataType(self.output_dtype)

        if isinstance(self.quantize_backend, str):
            self.quantize_backend = QuantizeBackend(self.quantize_backend)

        if isinstance(self.scale_rule, str):
            self.scale_rule = ScaleRule(self.scale_rule)

        if isinstance(self.weight_scale_rule, str):
            self.weight_scale_rule = ScaleRule(self.weight_scale_rule)

    def get_activation_scale_rule(self) -> ScaleRule:
        """Return the scaling rule to use for activation tensors."""
        return self.activation_scale_rule or self.scale_rule

    def get_gradient_scale_rule(self) -> ScaleRule:
        """Return the scaling rule to use for gradient tensors."""
        return self.gradient_scale_rule or self.scale_rule

    def get_weight_scale_rule(self) -> ScaleRule:
        """Return the scaling rule to use for weight tensors."""
        return self.weight_scale_rule or self.scale_rule

    def get_activation_config(self, **kwargs: dict[str, Any]) -> QuantizationConfig:
        """Return the quantization configuration for the activation tensors."""
        return QuantizationConfig(
            backend=self.quantize_backend,
            dtype=self.dtype,
            scale_rule=self.get_activation_scale_rule(),
            **kwargs,
        )

    def get_gradient_config(self, **kwargs: dict[str, Any]) -> QuantizationConfig:
        """Return the quantization configuration for the gradient tensors."""
        return QuantizationConfig(
            backend=self.quantize_backend,
            dtype=self.dtype,
            scale_rule=self.get_gradient_scale_rule(),
            **kwargs,
        )

    def get_weight_config(self, **kwargs: dict[str, Any]) -> QuantizationConfig:
        """Return the quantization configuration for the weight tensors."""
        return QuantizationConfig(
            backend=self.quantize_backend,
            dtype=self.dtype,
            scale_rule=self.get_weight_scale_rule(),
            **kwargs,
        )


@dataclass
class ModelQuantizationConfig(ModuleQuantizationConfig):
    """
    Configuration for quantizing a model with Four Over Six.

    Args:
        activation_scale_rule (ScaleRule | None): The scaling rule to use for activation
            tensors. If not provided, `scale_rule` will be used.
        dtype (DataType): The quantization data type to use for the module. Defaults to
            `DataType.nvfp4`.
        gradient_scale_rule (ScaleRule | None): The scaling rule to use for gradient
            tensors. If not provided, `scale_rule` will be used.
        keep_master_weights (bool): Whether to keep the master weights. Defaults to
            `False`.
        matmul_backend (MatmulBackend | None): The backend to use for matrix
            multiplications. If not provided, a backend will be selected automatically
            based on the available GPU and the specified options.
        output_dtype (DataType): The data type to use for the module's output. Defaults
            to `DataType.bfloat16`.
        quantize_backend (QuantizeBackend | None): The backend to use for quantization.
            If not provided, a backend will be selected automatically based on the
            available GPU and the specified options.
        scale_rule (ScaleRule): The fallback scaling rule which will be used if any of
            the other scaling rules are not specified.
        weight_scale_2d (bool): Whether to use 2D block scaling for weights. Should be
            set to `True` if the module is used for training.
        weight_scale_rule (ScaleRule | None): The scaling rule to use for weights. If
            not provided, `scale_rule` will be used.

        module_config_overrides (dict[str, ModuleQuantizationConfig]): A mapping of
            module names to quantization configurations to use for each module. If a
            module is not specified, the attributes from this class will be used.
        modules_to_not_convert (list[str]): A list of module names that should not be
            quantized.

    """

    module_config_overrides: dict[str, ModuleQuantizationConfig] = field(
        default_factory=dict,
    )
    modules_to_not_convert: list[str] = field(default_factory=lambda: ["lm_head"])

    def __post_init__(self) -> None:
        """Convert module config overrides to ModuleQuantizationConfig instances."""

        super().__post_init__()

        if self.module_config_overrides is not None:
            for module_name, module_config in self.module_config_overrides.items():
                if isinstance(module_config, dict):
                    self.module_config_overrides[module_name] = (
                        ModuleQuantizationConfig(**module_config)
                    )

    def get_module_config(self, module_name: str) -> ModuleQuantizationConfig:
        """Return the quantization configuration for a given module."""
        return (
            self.module_config_overrides.get(module_name, self)
            if self.module_config_overrides is not None
            else self
        )



from __future__ import annotations

import torch
from fouroversix.quantize.utils import to_blocked
from fouroversix.utils import DataType, RoundStyle, ScaleRule

E2M1_MAX_VALUE = 6
E2M1_MAX_FOUR = 4
E4M3_MAX_VALUE = 448
E4M3_MAX_FOUROVERSIX = 256


def fake_quantize_to_e2m1(
    x: torch.Tensor,
    *,
    round_style: RoundStyle = RoundStyle.nearest,
) -> torch.Tensor:
    if round_style == RoundStyle.nearest:
        step1 = torch.round(2 * x.abs()) / 2
        step2 = torch.round(x.abs())
        step3 = 2 * torch.round(x.abs() / 2)
    elif round_style == RoundStyle.stochastic:
        rbits = torch.rand_like(x.abs()) - 0.5
        step1 = torch.round(2 * x.abs() + rbits) / 2
        step2 = torch.round(x.abs() + rbits)
        step3 = 2 * torch.round(x.abs() / 2 + rbits)
        step3[step3 > E2M1_MAX_VALUE] = E2M1_MAX_VALUE

    mask1 = x.abs() < 2  # noqa: PLR2004
    mask2 = x.abs() < 4  # noqa: PLR2004

    return x.sign() * (
        step1 * mask1 + step2 * (~mask1) * mask2 + step3 * (~mask1) * (~mask2)
    )


def quantize_bf16_to_unpacked_fp4(x: torch.Tensor) -> torch.Tensor:
    assert x.dtype == torch.bfloat16

    bx = x.view(torch.int16)
    s = (bx >> 15) & 0x1
    e = (bx >> 7) & 0xFF
    m = bx & 0x7F
    is_zero = (e == 0) & (m == 0)

    # Default mantissa bit (for 1.5, 3.0, 6.0)
    m = (m >> 6) & 1
    is_half = (e == 126) & (m == 0)  # noqa: PLR2004
    m = torch.where(is_half, torch.tensor(1, dtype=torch.int16, device=x.device), m)

    # Exponent mapping
    # exp=126 -> E=0 (subnormals)
    # exp=127 -> E=1
    # exp=128 -> E=2
    # exp=129 -> E=3
    e = e - 126
    e = torch.where(is_zero, torch.tensor(0, dtype=torch.int16, device=x.device), e)

    # Zero always M=0
    m = torch.where(is_zero, torch.tensor(0, dtype=torch.int16, device=x.device), m)

    code = (s << 3) | (e << 1) | m
    return code.to(torch.uint8)


def pack_unpacked_fp4(x: torch.Tensor) -> torch.Tensor:
    assert x.dtype == torch.uint8

    dim = 1
    size_along_dim = x.size(dim)
    new_size_along_dim = (size_along_dim + 1) // 2

    # If the size is odd, we pad the data along dim with zeros at the end
    if size_along_dim % 2 != 0:
        pad_sizes = [0] * (2 * x.ndim)
        pad_index = (x.ndim - dim - 1) * 2 + 1
        pad_sizes[pad_index] = 1
        x = torch.nn.functional.pad(x, pad_sizes, mode="constant", value=0)

    new_shape = list(x.shape)
    new_shape[dim] = new_size_along_dim
    new_shape.insert(dim + 1, 2)  # packed dimension of length 2
    x = x.reshape(*new_shape)

    low = x.select(dim + 1, 0)
    high = x.select(dim + 1, 1)
    return (high << 4) | low


def quantize_to_mxfp4(
    x_scale_blocks: torch.Tensor,
    *,
    scale_rule: ScaleRule = ScaleRule.mse,
) -> tuple[torch.Tensor, torch.Tensor]:
    assert scale_rule in {ScaleRule.static_6, ScaleRule.static_4}

    x_scales_hp = (
        x_scale_blocks.abs().max(axis=-1).values / scale_rule.max_allowed_e2m1_value()
    )

    x_scales_e8m0_u32 = x_scales_hp.view(torch.int32)

    # Use the 8-bit exponent as the scale factor
    x_scales_e8m0 = ((x_scales_e8m0_u32 >> 23) & 0xFF).to(torch.uint8)

    # Add one in order to round up
    x_scales = torch.where(
        (x_scales_e8m0_u32 & 0x7FFFFF) == 0,
        x_scales_e8m0,
        x_scales_e8m0 + 1,
    )

    # Convert the rounded-up scale factor back to a 32-bit float
    x_scales_hp = (x_scales.to(torch.int32) << 23).view(torch.float32)

    x_block_scaled = x_scale_blocks / x_scales_hp.unsqueeze(1)

    return x_block_scaled, x_scales.view(torch.float8_e8m0fnu)


def quantize_to_nvfp4(
    x_scale_blocks: torch.Tensor,
    x_amax: torch.Tensor,
    *,
    scale_rule: ScaleRule,
    scale_expansion_factor: float | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    if x_amax == 0:
        x_scales_hp = torch.zeros(
            *x_scale_blocks.shape[:-1],
            dtype=x_amax.dtype,
            device=x_amax.device,
        )
    else:
        encode_scale = (
            torch.tensor(
                scale_rule.max_allowed_e2m1_value()
                * scale_rule.max_allowed_e4m3_value(),
                dtype=x_amax.dtype,
                device=x_amax.device,
            )
            / x_amax
        )
        x_scales_hp = (
            x_scale_blocks.abs().max(axis=-1).values
            / torch.tensor(
                scale_rule.max_allowed_e2m1_value(),
                dtype=x_amax.dtype,
                device=x_amax.device,
            )
            * encode_scale
        )

    if scale_expansion_factor is not None:
        x_scales_hp = x_scales_hp * scale_expansion_factor

    x_scales = x_scales_hp.to(torch.float8_e4m3fn)

    decode_scale = 1 / (
        torch.tensor(
            scale_rule.max_allowed_e2m1_value() * scale_rule.max_allowed_e4m3_value(),
            dtype=x_amax.dtype,
            device=x_amax.device,
        )
        / x_amax
    )
    x_block_scaled = torch.where(
        x_scales.unsqueeze(1) != 0,
        x_scale_blocks * (1 / (decode_scale * x_scales.to(x_amax.dtype).unsqueeze(1))),
        0,
    )

    return x_block_scaled, x_scales


def select_fouroversix(
    x_scale_blocks: torch.Tensor,
    x_block_scaled_6: torch.Tensor,
    scales_6: torch.Tensor,
    x_block_scaled_4: torch.Tensor,
    scales_4: torch.Tensor,
    x_amax: torch.Tensor,
    *,
    scale_rule: ScaleRule = ScaleRule.mse,
    round_style: RoundStyle = RoundStyle.nearest,
) -> tuple[torch.Tensor, torch.Tensor]:
    x_fake_quantized_6 = fake_quantize_to_e2m1(
        x_block_scaled_6,
        round_style=round_style,
    )
    x_fake_quantized_4 = fake_quantize_to_e2m1(
        x_block_scaled_4,
        round_style=round_style,
    )

    x_dequantized_6 = (
        x_fake_quantized_6.to(x_amax.dtype)
        * scales_6.unsqueeze(1).to(x_amax.dtype)
        * x_amax
        / torch.tensor(
            E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX,
            dtype=x_amax.dtype,
            device=x_amax.device,
        )
    )
    x_dequantized_4 = (
        x_fake_quantized_4.to(x_amax.dtype)
        * scales_4.unsqueeze(1).to(x_amax.dtype)
        * x_amax
        / torch.tensor(
            E2M1_MAX_VALUE * E4M3_MAX_FOUROVERSIX,
            dtype=x_amax.dtype,
            device=x_amax.device,
        )
    )

    if scale_rule == ScaleRule.abs_max:
        x_error_4 = (x_dequantized_4 - x_scale_blocks).abs().max(axis=-1).values
        x_error_6 = (x_dequantized_6 - x_scale_blocks).abs().max(axis=-1).values
    elif scale_rule == ScaleRule.mae:
        x_error_4 = (x_dequantized_4 - x_scale_blocks).abs().sum(axis=-1)
        x_error_6 = (x_dequantized_6 - x_scale_blocks).abs().sum(axis=-1)
    elif scale_rule == ScaleRule.mse:
        x_error_4 = ((x_dequantized_4 - x_scale_blocks) ** 2).sum(axis=-1)
        x_error_6 = ((x_dequantized_6 - x_scale_blocks) ** 2).sum(axis=-1)

    select_4 = (x_error_4 < x_error_6).unsqueeze(1)
    x_fake_quantized = torch.where(
        select_4,
        x_fake_quantized_4.reshape(x_scale_blocks.shape[0], -1),
        x_fake_quantized_6.reshape(x_scale_blocks.shape[0], -1),
    )
    scales = torch.where(
        select_4,
        scales_4.reshape(-1, 1),
        scales_6.reshape(-1, 1),
    )

    return x_fake_quantized, scales


def quantize_to_fp4(
    x: torch.Tensor,
    x_amax: torch.Tensor | None = None,
    had: torch.Tensor | None = None,
    *,
    block_scale_2d: bool = False,
    fp4_format: DataType = DataType.nvfp4,
    round_style: RoundStyle = RoundStyle.nearest,
    scale_rule: ScaleRule = ScaleRule.mse,
    transpose: bool = False,
) -> (
    tuple[torch.Tensor, torch.Tensor, torch.Tensor | None]
    | tuple[torch.Tensor, torch.Tensor, torch.Tensor | None, torch.Tensor]
):
    if transpose:
        x = x.T

    if had is not None:
        x = (x.reshape(-1, had.shape[0]) @ had).reshape_as(x)

    if x_amax is None:
        x_amax = (
            torch.ones(1, device=x.device, dtype=x.dtype)
            if fp4_format == DataType.mxfp4
            else x.abs().max().float()
        )

    if block_scale_2d:
        assert x.ndim == 2  # noqa: PLR2004
        assert x.shape[1] % fp4_format.block_size() == 0

        x_scale_blocks = (
            x.reshape(
                -1,
                fp4_format.block_size(),
                x.shape[1] // fp4_format.block_size(),
                fp4_format.block_size(),
            )
            .permute(0, 2, 1, 3)
            .reshape(-1, fp4_format.block_size() ** 2)
            .float()
        )
    else:
        x_scale_blocks = x.reshape(-1, fp4_format.block_size()).float()

    x_fake_quantized = None

    if fp4_format == DataType.mxfp4:
        x_block_scaled, scales = quantize_to_mxfp4(
            x_scale_blocks,
            scale_rule=scale_rule,
        )
    elif fp4_format == DataType.nvfp4 and scale_rule in {
        ScaleRule.static_6,
        ScaleRule.static_4,
    }:
        x_block_scaled, scales = quantize_to_nvfp4(
            x_scale_blocks,
            x_amax,
            scale_rule=scale_rule,
        )
    elif fp4_format == DataType.nvfp4:  # Four over six
        x_block_scaled_6, scales_6 = quantize_to_nvfp4(
            x_scale_blocks,
            x_amax,
            scale_rule=scale_rule,
        )
        x_block_scaled_4, scales_4 = quantize_to_nvfp4(
            x_scale_blocks,
            x_amax,
            scale_rule=scale_rule,
            scale_expansion_factor=1.5,
        )
        x_fake_quantized, scales = select_fouroversix(
            x_scale_blocks,
            x_block_scaled_6,
            scales_6,
            x_block_scaled_4,
            scales_4,
            x_amax,
            scale_rule=scale_rule,
            round_style=round_style,
        )
    else:
        msg = f"Invalid FP4 format: {fp4_format}"
        raise ValueError(msg)

    if x_fake_quantized is None:
        x_fake_quantized = fake_quantize_to_e2m1(
            x_block_scaled,
            round_style=round_style,
        )

    if block_scale_2d:
        x_fake_quantized = x_fake_quantized.reshape(
            -1,
            x.shape[1] // fp4_format.block_size(),
            fp4_format.block_size(),
            fp4_format.block_size(),
        ).permute(0, 2, 1, 3)

        scales = (
            scales.reshape(
                1,
                x.shape[0] // fp4_format.block_size(),
                x.shape[1] // fp4_format.block_size(),
            )
            .broadcast_to(
                fp4_format.block_size(),
                x.shape[0] // fp4_format.block_size(),
                x.shape[1] // fp4_format.block_size(),
            )
            .permute(1, 0, 2)
        )

    x_quantized = pack_unpacked_fp4(
        quantize_bf16_to_unpacked_fp4(x_fake_quantized.bfloat16().reshape_as(x)),
    )

    reshaped_scales = to_blocked(
        scales.reshape(
            x.shape[0],
            x.shape[1] // fp4_format.block_size(),
        ),
    )

    return x_quantized, reshaped_scales, x_amax



import functools

import torch
from fouroversix.quantize.backend import QuantizeBackendBase
from fouroversix.quantize.config import QuantizationConfig
from fouroversix.quantize.quantized_tensor import QuantizedTensor
from fouroversix.utils import BLACKWELL_SM_IDS, DataType, RoundStyle


class CUDAQuantizeBackend(QuantizeBackendBase):
    """
    The CUDA quantization backend. Supports basic quantization options (no 2D block
    scaling, no stochastic rounding, no random Hadamard transform). As a result, it can
    be used for inference, but not training. Requires a Blackwell GPU.
    """

    @classmethod
    @functools.lru_cache
    def is_available(cls) -> bool:
        """Return True if the CUDA backend is available on the current machine."""

        if (
            not torch.cuda.is_available()
            or torch.cuda.get_device_capability()[0] not in BLACKWELL_SM_IDS
        ):
            return False

        try:
            import fouroversix._C  # noqa: F401
        except ModuleNotFoundError:
            return False

        return True

    @classmethod
    def is_supported(cls, x: torch.Tensor, config: QuantizationConfig) -> bool:
        """
        Return True if the CUDA backend supports the given input and quantization
        configuration.
        """

        if not super().is_supported(x, config):
            return False

        return (
            x.device.type == "cuda"
            and config.round_style == RoundStyle.nearest
            and config.dtype == DataType.nvfp4
            and not config.transpose
        )

    @classmethod
    def quantize_to_fp4(
        cls,
        x: torch.Tensor,
        config: QuantizationConfig,
    ) -> QuantizedTensor:
        """
        Quantize a tensor to FP4 using the CUDA backend.

        Args:
            x (torch.Tensor): The input tensor to quantize.
            config (QuantizationConfig): The quantization configuration.

        Returns:
            The quantized tensor.

        """

        from .ops import quantize_to_fp4

        values, scale_factors, amax = quantize_to_fp4(
            x,
            config.dtype == DataType.nvfp4,
            config.round_style == RoundStyle.nearest,
            config.rht,
            config.block_scale_2d,
            config.transpose,
            config.scale_rule.cuda_id(),
            config.rbits,
        )

        return QuantizedTensor(
            values,
            scale_factors,
            amax,
            config.dtype,
            (x.shape[1], x.shape[0]) if config.transpose else x.shape,
            config.scale_rule,
        )



import torch
from fouroversix import DataType, ScaleRule
from fouroversix.quantize import QuantizedTensor
from transformers import ConversionOps, GptOssConfig, WeightConverter

from .conversions import WeightConversions


class FourOverSixGptOssDeserialize(ConversionOps):
    """Fouroversix deserializer for gpt oss model."""

    def __init__(
        self,
        dtype: DataType = None,
        scale_rule: ScaleRule = None,
    ) -> None:
        self.dtype = dtype
        self.scale_rule = scale_rule

    def convert(
        self,
        input_dict: torch.Tensor,
        **kwargs,  # noqa: ARG002, ANN003
    ) -> dict[str, list[torch.Tensor]]:
        """
        Convert the quantized parameters in gpt oss model to
        high precision weights.
        """

        prefix = ""
        if ".down_proj_blocks" in input_dict:
            weight = input_dict[".down_proj_blocks"][0]
            scales = input_dict[".down_proj_scales"][0]
            prefix = "down"
        elif ".gate_up_proj_blocks" in input_dict:
            weight = input_dict[".gate_up_proj_blocks"][0]
            scales = input_dict[".gate_up_proj_scales"][0]
            prefix = "gate_up"

        num_experts = weight.shape[0]
        hidden_size = weight.shape[1]
        weight = weight.reshape((num_experts, hidden_size, -1))

        dequantized_proj = []
        for e in range(num_experts):
            weight_uint8 = weight[e].to(torch.uint8)
            quantized_tensor = QuantizedTensor(
                values=weight_uint8,
                scale_factors=scales[e].to(torch.uint8).view(self.dtype.scale_dtype()),
                amax=torch.ones(
                    (1,),
                    device=weight[e].device,
                    dtype=torch.float32,
                ),
                dtype=self.dtype,
                original_shape=(
                    weight[e].shape[0],
                    weight[e].shape[1] * 2,
                ),
                scale_rule=self.scale_rule,
            )

            dequantized = quantized_tensor.dequantize()
            dequantized_proj.append(dequantized)

        dequantized_weight = torch.stack(dequantized_proj, dim=0)

        return {f"{prefix}_proj": [dequantized_weight]}

@WeightConversions.register(str(GptOssConfig))
class GptOssWeightConverter:
    """Stores the weight conversions for the gpt oss model."""

    @classmethod
    def get_weight_conversions(cls) -> list[WeightConverter]:
        """Return weight conversions for the gpt oss model."""
        return [
            WeightConverter(
                source_patterns=[".gate_up_proj_blocks", ".gate_up_proj_scales"],
                target_patterns=".gate_up_proj",
                operations=[FourOverSixGptOssDeserialize(
                    dtype=DataType.mxfp4,
                    scale_rule=ScaleRule.static_6,
                )],
            ),
            WeightConverter(
                source_patterns=[".down_proj_blocks", ".down_proj_scales"],
                target_patterns=".down_proj",
                operations=[FourOverSixGptOssDeserialize(
                    dtype=DataType.mxfp4,
                    scale_rule=ScaleRule.static_6,
                )],
            ),
        ]



import functools

import torch
from fouroversix.quantize.utils import to_blocked
from fouroversix.utils import BLACKWELL_SM_IDS, DataType, RoundStyle, ScaleRule

from .backend import QuantizeBackendBase
from .config import QuantizationConfig
from .quantized_tensor import QuantizedTensor


class TransformerEngineQuantizeBackend(QuantizeBackendBase):
    """
    A backend that quantizes inputs using NVIDIA's TransformerEngine. Used to debug
    our implementations.
    """

    @classmethod
    @functools.lru_cache
    def is_available(cls) -> bool:
        """
        Return True if the Transformer Engine backend is available on the current
        machine.
        """

        if (
            not torch.cuda.is_available()
            or torch.cuda.get_device_capability()[0] not in BLACKWELL_SM_IDS
        ):
            return False

        try:
            import transformer_engine  # noqa: F401
        except ModuleNotFoundError:
            return False

        return True

    @classmethod
    def is_supported(cls, x: torch.Tensor, config: QuantizationConfig) -> bool:
        """
        Return True if the Transformer Engine backend supports the given input and
        quantization configuration.
        """

        if not super().is_supported(x, config):
            return False

        if config.dtype != DataType.nvfp4 or config.scale_rule != ScaleRule.static_6:
            return False

        if not config.transpose and config.rht:
            return False

        if config.transpose and config.rht and config.block_scale_2d:  # noqa: SIM103
            return False

        return True

    @classmethod
    def quantize_to_fp4(
        cls,
        x: torch.Tensor,
        config: QuantizationConfig,
    ) -> QuantizedTensor:
        """
        Quantize a tensor to FP4 using the Transformer Engine backend.

        Args:
            x (torch.Tensor): The input tensor to quantize.
            config (QuantizationConfig): The quantization configuration.

        Returns:
            The quantized tensor.

        """

        from transformer_engine.pytorch.tensor.nvfp4_tensor import NVFP4Quantizer

        q = NVFP4Quantizer(
            with_2d_quantization=config.block_scale_2d,
            with_rht=config.rht,
            with_post_rht_amax=config.rht,
            stochastic_rounding=config.round_style == RoundStyle.stochastic,
        )

        out = q.quantize(x)

        if config.transpose:
            values = out._columnwise_data  # noqa: SLF001
            scale_factors = to_blocked(
                out._columnwise_scale_inv.view(torch.float8_e4m3fn),  # noqa: SLF001
            )
            amax = out._amax_columnwise  # noqa: SLF001
        else:
            values = out._rowwise_data  # noqa: SLF001
            scale_factors = to_blocked(
                out._rowwise_scale_inv.view(torch.float8_e4m3fn),  # noqa: SLF001
            )
            amax = out._amax_rowwise  # noqa: SLF001

        return QuantizedTensor(
            values,
            scale_factors,
            amax,
            config.dtype,
            (x.shape[1], x.shape[0]) if config.transpose else x.shape,
            config.scale_rule,
        )



import itertools

import pytest
import torch
from fouroversix import (
    DataType,
    QuantizationConfig,
    QuantizeBackend,
    RoundStyle,
    ScaleRule,
    quantize_to_fp4,
)
from fouroversix.quantize.frontend import AVAILABLE_BACKENDS
from fouroversix.quantize.quantized_tensor import from_blocked

MAE_MSE_MISMATCH_TOLERANCE = 1e-3
NUM_RANDOM_SEEDS = 10


@pytest.mark.parametrize("input_type", ["zeros", "ones", "rand01", "randn"])
@pytest.mark.parametrize(
    "input_shape",
    [(1024, 1024), (1024, 512), (512, 1024)],
)
@pytest.mark.parametrize(
    ("backend_a", "backend_b"),
    itertools.combinations(
        [
            QuantizeBackend.cuda,
            QuantizeBackend.triton,
            QuantizeBackend.pytorch,
            QuantizeBackend.transformer_engine,
        ],
        r=2,
    ),
)
@pytest.mark.parametrize("block_scale_2d", ["block_scale_2d", "no_block_scale_2d"])
@pytest.mark.parametrize("dtype", [DataType.nvfp4])
@pytest.mark.parametrize("rht", ["rht", "no_rht"])
@pytest.mark.parametrize(
    "scale_rule",
    [
        ScaleRule.abs_max,
        ScaleRule.mae,
        ScaleRule.mse,
        ScaleRule.static_4,
        ScaleRule.static_6,
    ],
)
@pytest.mark.parametrize("round_style", [RoundStyle.nearest, RoundStyle.stochastic])
@pytest.mark.parametrize("transpose", ["transpose", "no_transpose"])
def test_backend_outputs_are_consistent(  # noqa: C901, PLR0915
    input_type: str,
    input_shape: tuple[int, int],
    backend_a: QuantizeBackend,
    backend_b: QuantizeBackend,
    *,
    block_scale_2d: str,
    dtype: DataType,
    rht: str,
    round_style: RoundStyle,
    scale_rule: ScaleRule,
    transpose: str,
) -> None:
    torch.set_printoptions(precision=10)

    backend_a_cls = AVAILABLE_BACKENDS[backend_a]
    backend_b_cls = AVAILABLE_BACKENDS[backend_b]

    if not backend_a_cls.is_available() or not backend_b_cls.is_available():
        pytest.skip("Backend is not available")

    config_a = QuantizationConfig(
        backend=backend_a,
        block_scale_2d=block_scale_2d == "block_scale_2d",
        dtype=dtype,
        rht=rht == "rht",
        round_style=round_style,
        scale_rule=scale_rule,
        transpose=transpose == "transpose",
    )

    config_b = QuantizationConfig(
        backend=backend_b,
        block_scale_2d=block_scale_2d == "block_scale_2d",
        dtype=dtype,
        rht=rht == "rht",
        round_style=round_style,
        scale_rule=scale_rule,
        transpose=transpose == "transpose",
    )

    if round_style == RoundStyle.stochastic:
        pytest.xfail("This test is not currently targeting stochastic rounding")

    for random_seed in range(NUM_RANDOM_SEEDS):
        print(f"Testing with random seed: {random_seed}")
        torch.manual_seed(random_seed)

        if input_type == "zeros":
            x = torch.zeros(*input_shape, dtype=torch.bfloat16, device="cuda")
        elif input_type == "ones":
            x = torch.ones(*input_shape, dtype=torch.bfloat16, device="cuda")
        elif input_type == "rand01":
            x = torch.randint(0, 2, input_shape, dtype=int, device="cuda").to(
                torch.bfloat16,
            )
        elif input_type == "randn":
            x = torch.randn(*input_shape, dtype=torch.bfloat16, device="cuda")
        else:
            msg = f"Invalid input type: {input_type}"
            raise ValueError(msg)

        if not backend_a_cls.is_supported(
            x,
            config_a,
        ) or not backend_b_cls.is_supported(
            x,
            config_b,
        ):
            pytest.skip("Backend is not supported")

        quantized_a = quantize_to_fp4(x.clone(), config_a)
        quantized_b = quantize_to_fp4(x.clone(), config_b)

        if not torch.allclose(quantized_a.amax, quantized_b.amax):
            print("Backends A and B have different amax values!")
            print(f"{backend_a}: {quantized_a.amax}")
            print(f"{backend_b}: {quantized_b.amax}")
            pytest.fail("Backends A and B have different amax values!")

        sf_a = from_blocked(
            quantized_a.scale_factors.bfloat16(),
            (input_shape[0], input_shape[1] // 16),
        )
        sf_b = from_blocked(
            quantized_b.scale_factors.bfloat16(),
            (input_shape[0], input_shape[1] // 16),
        )

        # When computing 4/6 with the MAE and MSE scale rules, computing the errors
        # requires summing the errors in each block of 16 values. This operation
        # differently (elements are summed in different orders, and floating-point
        # addition is not associative) in PyTorch and Triton, and can not be easily made
        # deterministic in a way that allows for good performance. As a result, we allow
        # a small number of mismatches between the scale factors and values for these
        # two rules. Fortunately, abs_max does not involve a summation, so we can use it
        # to test the correctness of the rest of the 4/6 implementation.
        scale_factors_mismatch_prop = (sf_a != sf_b).sum() / sf_a.numel()

        if (
            scale_rule in {ScaleRule.static_6, ScaleRule.static_4, ScaleRule.abs_max}
            and scale_factors_mismatch_prop > 0
        ) or scale_factors_mismatch_prop >= MAE_MSE_MISMATCH_TOLERANCE:
            print(
                "Backends A and B have different scale factors! "
                f"{scale_factors_mismatch_prop:.2%} mismatch",
            )

            [i, *_], [j, *_] = torch.where(sf_a != sf_b)
            print(backend_a)
            print("sf", sf_a[i, j])
            print("e2m1", quantized_a.values[i, 8 * j : 8 * (j + 1)])
            print(backend_b)
            print("sf", sf_b[i, j])
            print("e2m1", quantized_b.values[i, 8 * j : 8 * (j + 1)])
            print("original")
            print("x", x[i, 16 * j : 16 * (j + 1)])
            pytest.fail("Backends A and B have different scale factors!")

        values_mismatch_prop = (
            quantized_a.values != quantized_b.values
        ).sum() / quantized_a.values.numel()

        if (
            scale_rule in {ScaleRule.static_6, ScaleRule.static_4, ScaleRule.abs_max}
            and values_mismatch_prop > 0
        ) or values_mismatch_prop >= MAE_MSE_MISMATCH_TOLERANCE:
            print(
                "Backends A and B have different e2m1 values! "
                f"{values_mismatch_prop:.2%} mismatch",
            )

            [i, *_], [j, *_] = torch.where(
                quantized_a.values != quantized_b.values,
            )
            print(i, j)
            print("amax", quantized_a.amax)
            print("sf", sf_a[i, j // 8])
            print(backend_a)
            print("e2m1", quantized_a.values[i, 8 * (j // 8) : 8 * (j // 8 + 1)])
            print(backend_b)
            print("e2m1", quantized_b.values[i, 8 * (j // 8) : 8 * (j // 8 + 1)])
            print("original")
            print("x", x[i, 16 * (j // 8) : 16 * (j // 8 + 1)])
            pytest.fail("Backends A and B have different e2m1 values!")



import functools

import torch
from fouroversix.quantize.backend import QuantizeBackendBase
from fouroversix.quantize.config import QuantizationConfig
from fouroversix.quantize.quantized_tensor import QuantizedTensor
from fouroversix.quantize.utils import get_rht_matrix
from fouroversix.utils import BLACKWELL_SM_IDS, SM_100, RoundStyle


class TritonQuantizeBackend(QuantizeBackendBase):
    """
    The Triton quantization backend. Supports all parameters required for efficient
    NVFP4 training, including stochastic rounding, the random Hadamard transform,
    transposed inputs, and 2D block scaling. Requires a Blackwell GPU.
    """

    @classmethod
    @functools.lru_cache
    def is_available(cls) -> bool:
        """Return True if the Triton backend is available on the current machine."""
        return (
            torch.cuda.is_available()
            and torch.cuda.get_device_capability()[0] in BLACKWELL_SM_IDS
        )

    @classmethod
    def is_supported(cls, x: torch.Tensor, config: QuantizationConfig) -> bool:
        """
        Return True if the Triton backend supports the given input and quantization
        configuration.
        """

        if not super().is_supported(x, config):
            return False

        if config.round_style == RoundStyle.stochastic:
            return torch.cuda.get_device_capability()[0] == SM_100

        return x.device.type == "cuda"

    @classmethod
    def quantize_to_fp4(
        cls,
        x: torch.Tensor,
        config: QuantizationConfig,
    ) -> QuantizedTensor:
        """
        Quantize a tensor to FP4 using the Triton backend.

        Args:
            x (torch.Tensor): The input tensor to quantize.
            config (QuantizationConfig): The quantization configuration.

        Returns:
            The quantized tensor.

        """

        from .kernel import quantize_to_fp4

        values, scale_factors, amax = quantize_to_fp4(
            x,
            had=get_rht_matrix() if config.rht else None,
            fp4_format=config.dtype,
            round_style=config.round_style,
            scale_rule=config.scale_rule,
            block_scale_2d=config.block_scale_2d,
            transpose=config.transpose,
            rbits=config.rbits,
        )

        return QuantizedTensor(
            values,
            scale_factors,
            amax,
            config.dtype,
            (x.shape[1], x.shape[0]) if config.transpose else x.shape,
            config.scale_rule,
        )



from typing import Any

import torch
import torch.nn as nn
from fouroversix.matmul import fp4_matmul
from fouroversix.model.config import ModuleQuantizationConfig
from fouroversix.model.quantize import QuantizedModule
from fouroversix.quantize import (
    QuantizationConfig,
    QuantizedTensor,
    quantize_to_fp4,
)
from fouroversix.utils import RoundStyle


class FourOverSixLinearFunction(torch.autograd.Function):
    """Differentiable FP4 linear layer."""

    @staticmethod
    def forward(
        ctx: torch.autograd.function.FunctionCtx,
        config: ModuleQuantizationConfig,
        input: torch.Tensor,
        weight: torch.Tensor | QuantizedTensor,
        bias: torch.Tensor = None,
    ) -> tuple[torch.Tensor,]:
        """
        Perform an FP4 matrix multiplication. The input is provided in high precision
        and quantized to FP4 prior to the matrix multiplication, while the weight is
        provided in low precision.
        """

        fprop_activation_config = config.get_activation_config()
        fprop_weight_config = config.get_weight_config(
            block_scale_2d=config.weight_scale_2d,
        )

        if isinstance(weight, torch.Tensor):
            ctx.save_for_backward(input, weight, bias)
            weight = quantize_to_fp4(weight, fprop_weight_config)

        ctx.config = config

        out = fp4_matmul(
            input.reshape(-1, input.shape[-1]),
            weight,
            backend=config.matmul_backend,
            input_config=fprop_activation_config,
            out_dtype=config.output_dtype,
        ).reshape(*input.shape[:-1], weight.original_shape[0])

        if bias is not None:
            out = out + bias

        return out

    @staticmethod
    def backward(
        ctx: torch.autograd.function.FunctionCtx,
        grad_output: torch.Tensor,
    ) -> tuple[torch.Tensor, ...]:
        """Backward pass for the FP4 linear layer."""

        input, weight, bias = ctx.saved_tensors

        assert grad_output.shape[0] == 1  # noqa: S101

        dgrad_grad_config = ctx.config.get_gradient_config(
            round_style=RoundStyle.stochastic,
        )

        dgrad_weight_config = ctx.config.get_weight_config(
            block_scale_2d=ctx.config.weight_scale_2d,
            transpose=True,
        )

        grad_input = fp4_matmul(
            grad_output[0],
            weight,
            backend=ctx.matmul_backend,
            input_config=dgrad_grad_config,
            other_config=dgrad_weight_config,
            out_dtype=ctx.config.output_dtype,
        ).unsqueeze(0)

        wgrad_grad_config = ctx.config.get_gradient_config(
            rht=True,
            round_style=RoundStyle.stochastic,
            transpose=True,
        )

        wgrad_activation_config = ctx.config.get_activation_config(
            rht=True,
            transpose=True,
        )

        grad_weight = fp4_matmul(
            grad_output[0],
            input[0],
            backend=ctx.matmul_backend,
            input_config=wgrad_grad_config,
            other_config=wgrad_activation_config,
            out_dtype=ctx.config.output_dtype,
        ).unsqueeze(0)

        grad_bias = (
            grad_output.sum(0) if bias is not None and ctx.needs_input_grad[3] else None
        )

        return (
            None,
            grad_input,
            grad_weight,
            grad_bias,
        )


@QuantizedModule.register(nn.Linear)
class FourOverSixLinear(nn.Linear):
    """
    Drop-in replacement for `nn.Linear` that quantizes weights, activations, and
    gradients.
    """

    def __init__(
        self,
        module: nn.Linear,
        config: ModuleQuantizationConfig,
    ) -> None:
        """
        Initialize the FourOverSixLinear layer.

        Args:
            module (nn.Linear): The high-precision module that this quantized layer will
                replace.
            config (ModuleQuantizationConfig): The quantization configuration to use for
                the layer.

        """

        super().__init__(
            module.in_features,
            module.out_features,
            module.bias is not None,
            module.weight.device,
            module.weight.dtype,
        )

        self.weight = module.weight
        self.bias = module.bias
        self.config = config

        if not self.config.keep_master_weights:
            self.register_buffer(
                "quantized_weight_values",
                nn.Parameter(
                    torch.zeros(
                        self.out_features,
                        self.in_features // 2,
                        dtype=torch.uint8,
                    ),
                    requires_grad=False,
                ),
            )
            self.register_buffer(
                "quantized_weight_scale_factors",
                nn.Parameter(
                    torch.zeros(
                        self.out_features
                        * self.in_features
                        // self.config.dtype.block_size(),
                        dtype=self.config.dtype.scale_dtype(),
                    ),
                    requires_grad=False,
                ),
            )
            self.register_buffer(
                "quantized_weight_amax",
                nn.Parameter(torch.zeros(1, dtype=torch.float32), requires_grad=False),
            )
            self.register_buffer(
                "quantized_weight_metadata",
                nn.Parameter(
                    torch.zeros(2 + 2, dtype=torch.int32),
                    requires_grad=False,
                ),
            )

    @property
    def parameters_to_quantize(self) -> tuple[str, ...]:
        """Return high precision parameters to be quantized and deleted."""
        return ("weight",)

    def get_packing_factor(self, parameter_name: str) -> float:
        """Get the packing factor for a parameter."""
        return 2 if parameter_name == "quantized_weight_values" else 1

    def get_quantized_parameters(
        self,
        parameter_name: str,
        parameter: torch.Tensor,
    ) -> dict[str, Any]:
        """Get the quantized parameters for the layer."""

        if parameter_name == "weight":
            config = QuantizationConfig(
                backend=self.config.quantize_backend,
                block_scale_2d=self.config.weight_scale_2d,
                dtype=self.config.dtype,
                scale_rule=self.config.get_weight_scale_rule(),
            )

            quantized_weight = quantize_to_fp4(parameter, config)

            return {
                "quantized_weight_values": quantized_weight.values,
                "quantized_weight_scale_factors": quantized_weight.scale_factors,
                "quantized_weight_amax": quantized_weight.amax,
                "quantized_weight_metadata": torch.tensor(
                    [
                        quantized_weight.original_shape[0],
                        quantized_weight.original_shape[1],
                        quantized_weight.padded_shape[0],
                        quantized_weight.padded_shape[1],
                    ],
                ),
            }

        msg = f"Unsupported high-preciison parameter: {parameter_name}"
        raise ValueError(msg)

    def quantized_weight(self) -> QuantizedTensor:
        """
        Prepare this layer for post-training quantization by quantizing the weight,
        storing the quantized weight, and deleting the original weight. This should not
        be done if the layer is used for training, as training requires storage of the
        high-precision weight.
        """

        if not hasattr(self, "_quantized_weight"):
            if self.config.keep_master_weights:
                return quantize_to_fp4(self.weight, self.config.get_weight_config())

            original_shape = tuple(self.quantized_weight_metadata.data[:2].tolist())
            padded_shape = tuple(self.quantized_weight_metadata.data[2:].tolist())

            self._quantized_weight = QuantizedTensor(
                self.quantized_weight_values.data,
                self.quantized_weight_scale_factors.data,
                self.quantized_weight_amax.data,
                self.config.dtype,
                original_shape,
                self.config.get_weight_scale_rule(),
                padded_shape,
            )

        return self._quantized_weight

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Forward pass for the FP4 linear layer."""

        return FourOverSixLinearFunction.apply(
            self.config,
            input,
            self.quantized_weight(),
            self.bias,
        )



import torch
from fouroversix.quantize import QuantizationConfig, QuantizedTensor, quantize_to_fp4
from fouroversix.utils import DataType, MatmulBackend

from .cutlass import CUTLASSMatmulBackend
from .pytorch import PyTorchMatmulBackend

AVAILABLE_BACKENDS = {
    MatmulBackend.cutlass: CUTLASSMatmulBackend,
    MatmulBackend.pytorch: PyTorchMatmulBackend,
}


def fp4_matmul(
    input: torch.Tensor | QuantizedTensor,
    other: torch.Tensor | QuantizedTensor,
    *,
    backend: MatmulBackend | None = None,
    input_config: QuantizationConfig | None = None,
    other_config: QuantizationConfig | None = None,
    out_dtype: DataType = DataType.bfloat16,
) -> torch.Tensor:
    """
    Perform a matrix multiplication (`a @ b.T`) between two quantized tensors.

    ## Sample Code

    Each tensor may be provided in either high or low precision. If provided in high
    precision, tensors will be quantized to FP4 prior to the matrix multiplication, and
    quantization may be configured with the `input_quantize_kwargs` and
    `other_quantize_kwargs` parameters. For example, the following two code samples are
    equivalent:

    ### With High-Precision Inputs

    ```python
    a = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    b = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    out = fp4_matmul(a, b)
    ```

    ### With Low-Precision Inputs

    ```python
    a = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")
    b = torch.tensor(1024, 1024, dtype=torch.bfloat16, device="cuda")

    a_quantized = quantize_to_fp4(a)
    b_quantized = quantize_to_fp4(b)

    out = fp4_matmul(a_quantized, b_quantized)
    ```

    ## Backends

    We provide two different implementations of FP4 matrix multiplication:

    - **CUTLASS**: Uses CUTLASS kernels to perform fast FP4 matrix multiplication.
        Requires a Blackwell GPU.
    - **PyTorch**: A slow implementation which dequantizes FP4 tensors and then
        performs a high-precision matrix multiplication.

    ## Parameters

    Args:
        input (torch.Tensor | QuantizedTensor): The first tensor to be multiplied.
        other (torch.Tensor | QuantizedTensor): The second tensor to be multiplied.
        backend (MatmulBackend): The backend to use for the matrix multiplication,
            either `MatmulBackend.cutlass` or `MatmulBackend.pytorch`. If no backend is
            provided, CUTLASS will be used if the machine has a Blackwell GPU, and
            PyTorch will be used otherwise.
        input_config (QuantizationConfig | None): If `input` is provided in high
            precision, this configuration will be passed to the `quantize_to_fp4` call
            done prior to the matrix multiplication.
        other_config (QuantizationConfig | None): If `other` is provided in high
            precision, this configuration will be passed to the `quantize_to_fp4` call
            done prior to the matrix multiplication.
        out_dtype (DataType): The data type of the output tensor. Defaults to
            `DataType.bfloat16`.

    Returns:
        The output tensor.

    """

    if input_config is None:
        input_config = QuantizationConfig()

    if isinstance(input, torch.Tensor):
        input = quantize_to_fp4(input, input_config)

    if other_config is None:
        other_config = QuantizationConfig()

    if isinstance(other, torch.Tensor):
        other = quantize_to_fp4(other, other_config)

    if backend is None:
        for backend_candidate in [MatmulBackend.cutlass, MatmulBackend.pytorch]:
            if AVAILABLE_BACKENDS[backend_candidate] is not None and AVAILABLE_BACKENDS[
                backend_candidate
            ].is_supported(input, other, out_dtype=out_dtype):
                backend = backend_candidate
                break
        else:
            msg = "No backend found that supports the given parameters"
            raise ValueError(msg)

    elif not AVAILABLE_BACKENDS[backend].is_supported(
        input, other, out_dtype=out_dtype,
    ):
        msg = f"Backend {backend} does not support the given parameters"
        raise ValueError(msg)

    return AVAILABLE_BACKENDS[backend].fp4_matmul(input, other, out_dtype=out_dtype)



import torch

import fouroversix._C  # noqa: F401


def quantize_to_fp4(
    x: torch.Tensor,
    is_nvfp4: bool,  # noqa: FBT001
    is_rtn: bool,  # noqa: FBT001
    is_rht: bool,  # noqa: FBT001
    is_2d: bool,  # noqa: FBT001
    is_transpose: bool,  # noqa: FBT001
    selection_rule: int,
    rbits: int,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return torch.ops.fouroversix.quantize_to_fp4.default(
        x,
        is_nvfp4,
        is_rtn,
        is_rht,
        is_2d,
        is_transpose,
        selection_rule,
        rbits,
    )


@torch.library.register_fake("fouroversix::quantize_to_fp4")
def _(
    x: torch.Tensor,
    is_nvfp4: bool,  # noqa: FBT001
    is_rtn: bool,  # noqa: ARG001, FBT001
    is_rht: bool,  # noqa: ARG001, FBT001
    is_transpose: bool,  # noqa: ARG001, FBT001
    selection_rule: int,  # noqa: ARG001
    rbits: int,  # noqa: ARG001
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    return (
        torch.empty(x.shape[0], x.shape[1] // 2, dtype=torch.uint8, device=x.device),
        torch.empty(
            x.shape[0] * x.shape[1] // (16 if is_nvfp4 else 32),
            dtype=torch.float8_e4m3fn if is_nvfp4 else torch.float8_e8m0fnu,
            device=x.device,
        ),
        torch.empty(1, dtype=torch.float32, device=x.device),
    )



import functools

import torch
from fouroversix.matmul.backend import MatmulBackendBase
from fouroversix.quantize import QuantizedTensor
from fouroversix.utils import BLACKWELL_SM_IDS, SM_100, SM_120, DataType


class CUTLASSMatmulBackend(MatmulBackendBase):
    """
    The CUTLASS matrix multiplication backend. Uses CUTLASS kernels to perform fast
    FP4 matrix multiplication. Requires a Blackwell GPU.
    """

    @classmethod
    @functools.lru_cache
    def is_available(cls) -> bool:
        """Return True if the CUTLASS backend is available on the current machine."""

        if (
            not torch.cuda.is_available()
            or torch.cuda.get_device_capability()[0] not in BLACKWELL_SM_IDS
        ):
            return False

        try:
            import fouroversix._C  # noqa: F401
        except ModuleNotFoundError:
            return False

        return True

    @classmethod
    def is_supported(
        cls,
        input: QuantizedTensor,
        other: QuantizedTensor,
        *,
        out_dtype: DataType,
    ) -> bool:
        """
        Return True if the CUTLASS backend supports the given inputs and output data
        type.
        """

        if not super().is_supported(input, other, out_dtype=out_dtype):
            return False

        return input.device.type == "cuda"

    @classmethod
    def fp4_matmul(
        cls,
        input: QuantizedTensor,
        other: QuantizedTensor,
        *,
        out_dtype: DataType,
    ) -> torch.Tensor:
        """
        Perform a matrix multiplication (`a @ b.T`) between two quantized tensors using
        the CUTLASS backend.
        """

        from .ops import (
            gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt,
            gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120,
            gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt,
            gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120,
            gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt,
            gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120,
        )

        out_shape = (input.original_shape[0], other.original_shape[0])

        if input.dtype == DataType.mxfp4:
            alpha = torch.ones(
                1,
                device=input.values.device,
                dtype=torch.float32,
            )
        elif input.dtype == DataType.nvfp4:
            alpha = (
                (input.amax * other.amax)
                / (
                    input.scale_rule.max_allowed_e2m1_value()
                    * input.scale_rule.max_allowed_e4m3_value()
                    * other.scale_rule.max_allowed_e2m1_value()
                    * other.scale_rule.max_allowed_e4m3_value()
                )
            ).to(torch.float32)

        gemm_fns = {
            (
                SM_100,
                DataType.mxfp4,
                DataType.bfloat16,
            ): gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt,
            (
                SM_120,
                DataType.mxfp4,
                DataType.bfloat16,
            ): gemm_mxfp4mxfp4_accum_fp32_out_bf16_tnt_sm120,
            (
                SM_100,
                DataType.nvfp4,
                DataType.bfloat16,
            ): gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt,
            (
                SM_120,
                DataType.nvfp4,
                DataType.bfloat16,
            ): gemm_nvfp4nvfp4_accum_fp32_out_bf16_tnt_sm120,
            (
                SM_100,
                DataType.nvfp4,
                DataType.float16,
            ): gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt,
            (
                SM_120,
                DataType.nvfp4,
                DataType.float16,
            ): gemm_nvfp4nvfp4_accum_fp32_out_fp16_tnt_sm120,
        }

        gemm_fn = gemm_fns.get(
            (torch.cuda.get_device_capability()[0], input.dtype, out_dtype),
        )

        if gemm_fn is None:
            msg = (
                "No gemm function found for the given device capability and "
                f"out_dtype: {torch.cuda.get_device_capability()[0]}, {out_dtype}"
            )
            raise ValueError(msg)

        out = gemm_fn(
            input.values,
            other.values,
            input.scale_factors,
            other.scale_factors,
            alpha,
        )

        if out_shape is not None and out.shape != out_shape:
            out = out[: out_shape[0], : out_shape[1]]

        return out



import warnings
from typing import Any

import torch
from fouroversix.matmul import fp4_matmul
from fouroversix.model.config import ModuleQuantizationConfig
from fouroversix.model.quantize import QuantizedModule
from fouroversix.quantize import QuantizationConfig, QuantizedTensor, quantize_to_fp4
from torch import nn

try:
    from transformers.models.qwen3_5_moe.modeling_qwen3_5_moe import Qwen3_5MoeExperts
except ImportError:
    warnings.warn(
        "Qwen3_5MoeExperts not found, please update transformers to the latest "
        "version to quantize Qwen3.5 models.",
        stacklevel=2,
    )

    Qwen3_5MoeExperts = None


@QuantizedModule.register(Qwen3_5MoeExperts)
class FourOverSixQwenExperts(nn.Module):
    """
    Drop-in replacement for the Qwen3_5MoeExperts layer that
    uses FP4 quantization.
    """

    def __init__(
        self,
        module: Qwen3_5MoeExperts,
        config: ModuleQuantizationConfig,
    ) -> None:
        """
        Initialize the FourOverSixQwenExperts layer.

        Args:
            module (GptOssMLP): The high-precision module that this quantized layer will
                replace.
            config (ModuleQuantizationConfig): The quantization configuration to use for
                the layer.

        """
        super().__init__()

        self.num_experts = module.num_experts
        self.intermediate_dim = module.intermediate_dim
        self.hidden_dim = module.hidden_dim

        self.down_proj = module.down_proj
        self.gate_up_proj = module.gate_up_proj

        self.device = self.down_proj.device
        self.config = config

        self.act_fn = module.act_fn

        if not self.config.keep_master_weights:

            self.register_buffer(
                "quantized_down_proj_values",
                nn.Parameter(
                    torch.zeros(
                        self.num_experts,
                        self.hidden_dim,
                        self.intermediate_dim // 2,
                        dtype=torch.uint8,
                    ),
                    requires_grad=False,
                ),
            )

            self.register_buffer(
                "quantized_gate_up_proj_values",
                nn.Parameter(
                    torch.zeros(
                        self.num_experts,
                        self.intermediate_dim * 2,
                        self.hidden_dim // 2,
                        dtype=torch.uint8,
                    ),
                    requires_grad=False,
                ),
            )

            self.register_buffer(
                "quantized_down_proj_scale_factors",
                nn.Parameter(
                    torch.zeros(
                        self.num_experts,
                        self.hidden_dim
                        * self.intermediate_dim
                        // self.config.dtype.block_size(),
                        dtype=self.config.dtype.scale_dtype(),
                    ),
                    requires_grad=False,
                ),
            )
            self.register_buffer(
                "quantized_gate_up_proj_scale_factors",
                nn.Parameter(
                    torch.zeros(
                        self.num_experts,
                        self.hidden_dim
                        * (self.intermediate_dim * 2)
                        // self.config.dtype.block_size(),
                        dtype=self.config.dtype.scale_dtype(),
                    ),
                    requires_grad=False,
                ),
            )

            self.register_buffer(
                "quantized_down_proj_amax",
                nn.Parameter(
                    torch.zeros(self.num_experts, 1, dtype=torch.float32),
                    requires_grad=False,
                ),
            )
            self.register_buffer(
                "quantized_gate_up_proj_amax",
                nn.Parameter(
                    torch.zeros(self.num_experts, 1, dtype=torch.float32),
                    requires_grad=False,
                ),
            )

            self.register_buffer(
                "quantized_down_proj_metadata",
                nn.Parameter(
                    torch.zeros(self.num_experts, 4, dtype=torch.int32),
                    requires_grad=False,
                ),
            )
            self.register_buffer(
                "quantized_gate_up_proj_metadata",
                nn.Parameter(
                    torch.zeros(self.num_experts, 4, dtype=torch.int32),
                    requires_grad=False,
                ),
            )

    @property
    def parameters_to_quantize(self) -> tuple[str, ...]:
        """Return high precision parameters to be quantized and deleted."""
        return ("down_proj", "gate_up_proj")

    def get_packing_factor(self, parameter_name: str) -> float:
        """Get the packing factor for a parameter."""
        return (
            2
            if parameter_name
            in {"quantized_down_proj_values", "quantized_gate_up_proj_values"}
            else 1
        )

    def get_quantized_parameters(
        self,
        parameter_name: str,
        parameter: torch.Tensor,
    ) -> dict[str, Any]:
        """
        Prepare this layer for post-training quantization by quantizing the weight,
        storing the quantized weight, and deleting the original weight. This should not
        be done if the layer is used for training, as training requires storage of the
        high-precision weight.
        """

        if "bias" in parameter_name:
            return {parameter_name: parameter}

        weight_config = QuantizationConfig(
            backend=self.config.quantize_backend,
            dtype=self.config.dtype,
            scale_rule=self.config.get_weight_scale_rule(),
        )

        quantized_proj = []
        for e in range(parameter.shape[0]):
            q = quantize_to_fp4(parameter[e], weight_config)
            quantized_proj.append(q)

        if "down" in parameter_name:
            prefix = "down"
        elif "gate_up" in parameter_name:
            prefix = "gate_up"

        return {
            f"quantized_{prefix}_proj_values": torch.stack(
                [tensor.values for tensor in quantized_proj],
                dim=0,
            ),
            f"quantized_{prefix}_proj_scale_factors": torch.stack(
                [tensor.scale_factors for tensor in quantized_proj],
                dim=0,
            ),
            f"quantized_{prefix}_proj_amax": torch.stack(
                [tensor.amax for tensor in quantized_proj],
                dim=0,
            ),
            f"quantized_{prefix}_proj_metadata": torch.stack(
                [
                    torch.tensor(
                        [
                            tensor.original_shape[0],
                            tensor.original_shape[1],
                            tensor.padded_shape[0],
                            tensor.padded_shape[1],
                        ],
                    )
                    for tensor in quantized_proj
                ],
            ),
        }

    def forward(
        self,
        hidden_states: torch.Tensor,
        top_k_index: torch.Tensor,
        top_k_weights: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass for the FP4 experts layer."""

        down_proj, gate_up_proj = self.quantized_weights()

        final_hidden_states = torch.zeros_like(hidden_states)
        with torch.no_grad():
            expert_mask = torch.nn.functional.one_hot(
                top_k_index,
                num_classes=self.num_experts,
            )
            expert_mask = expert_mask.permute(2, 1, 0)
            expert_hit = torch.greater(expert_mask.sum(dim=(-1, -2)), 0).nonzero()

        for [expert_idx] in expert_hit:
            if expert_idx == self.num_experts:
                continue

            fprop_activation_config = QuantizationConfig(
                backend=self.config.quantize_backend,
                dtype=self.config.dtype,
                scale_rule=self.config.get_activation_scale_rule(),
            )
            top_k_pos, token_idx = torch.where(expert_mask[expert_idx])
            current_state = hidden_states[token_idx]
            gate, up = fp4_matmul(
                current_state,
                gate_up_proj[expert_idx],
                input_config=fprop_activation_config,
                out_dtype=self.config.output_dtype,
            ).chunk(2, dim=-1)
            current_hidden_states = self.act_fn(gate) * up
            current_hidden_states = fp4_matmul(
                current_hidden_states,
                down_proj[expert_idx],
                input_config=fprop_activation_config,
                out_dtype=self.config.output_dtype,
            )
            current_hidden_states = (
                current_hidden_states * top_k_weights[token_idx, top_k_pos, None]
            )
            final_hidden_states.index_add_(
                0,
                token_idx,
                current_hidden_states.to(final_hidden_states.dtype),
            )

        return final_hidden_states

    def quantized_weights(self) -> tuple[list[QuantizedTensor], list[QuantizedTensor]]:
        """Return quantized parameters as QuantizedTensor."""

        if not hasattr(self, "_quantized_weights"):
            weight_config = self.config.get_weight_config()
            if self.config.keep_master_weights:
                down = [
                    quantize_to_fp4(self.down_proj[e], weight_config)
                    for e in range(self.num_experts)
                ]
                gate_up = [
                    quantize_to_fp4(self.gate_up_proj[e], weight_config)
                    for e in range(self.num_experts)
                ]
                return (down, gate_up)

            down = []
            gate_up = []
            for e in range(self.num_experts):
                down.append(
                    QuantizedTensor(
                        values=self.quantized_down_proj_values.data[e],
                        scale_factors=self.quantized_down_proj_scale_factors.data[e],
                        amax=self.quantized_down_proj_amax.data[e],
                        dtype=self.config.dtype,
                        original_shape=tuple(
                            self.quantized_down_proj_metadata.data[e, :2].tolist(),
                        ),
                        scale_rule=self.config.get_weight_scale_rule(),
                        padded_shape=tuple(
                            self.quantized_down_proj_metadata.data[e, 2:].tolist(),
                        ),
                    ),
                )
                gate_up.append(
                    QuantizedTensor(
                        values=self.quantized_gate_up_proj_values.data[e],
                        scale_factors=self.quantized_gate_up_proj_scale_factors.data[e],
                        amax=self.quantized_gate_up_proj_amax.data[e],
                        dtype=self.config.dtype,
                        original_shape=tuple(
                            self.quantized_gate_up_proj_metadata.data[e, :2].tolist(),
                        ),
                        scale_rule=self.config.get_weight_scale_rule(),
                        padded_shape=tuple(
                            self.quantized_gate_up_proj_metadata.data[e, 2:].tolist(),
                        ),
                    ),
                )
            self._quantized_weights = (down, gate_up)

        return self._quantized_weights



import functools
import os
import platform
import subprocess
import sys
import urllib.request
import warnings
from pathlib import Path
from typing import Any

import torch
from packaging.version import Version, parse
from setuptools import setup
from setuptools.command.bdist_wheel import bdist_wheel
from torch.utils.cpp_extension import CUDA_HOME, BuildExtension, CUDAExtension

BASE_WHEEL_URL = "https://github.com/mit-han-lab/fouroversix/releases/download"
PACKAGE_NAME = "fouroversix"
PACKAGE_VERSION = "1.0.4"

CUTLASS_DEBUG = os.getenv("CUTLASS_DEBUG", "0") == "1"
FORCE_BUILD = os.getenv("FORCE_BUILD", "0") == "1"
FORCE_CXX11_ABI = os.getenv("FORCE_CXX11_ABI", "0") == "1"
SKIP_CUDA_BUILD = os.getenv("SKIP_CUDA_BUILD", "0") == "1"


@functools.cache
def get_cuda_archs() -> list[str]:
    return os.getenv("CUDA_ARCHS", "100;103;110;120").split(";")


def get_cuda_bare_metal_version() -> Version | None:
    if CUDA_HOME is None:
        warnings.warn(
            "nvcc was not found. Are you sure your environment has nvcc available? If "
            "you're installing within a container from "
            "https://hub.docker.com/r/pytorch/pytorch, only images with 'devel' in "
            "their name will provide nvcc.",
            stacklevel=1,
        )
        return None

    raw_output = subprocess.check_output(
        [CUDA_HOME + "/bin/nvcc", "-V"],
        universal_newlines=True,
    )
    output = raw_output.split()
    release_idx = output.index("release") + 1
    return parse(output[release_idx].split(",")[0])


def get_cuda_gencodes() -> list[str]:
    """
    Add -gencode flags based on nvcc capabilities.

    Uses the following rules:
      - sm_100/120 on CUDA >= 12.8
      - Use 100f on CUDA >= 12.9 (Blackwell family-specific)
      - Map requested 110 -> 101 if CUDA < 13.0 (Thor rename)
      - Embed PTX for newest arch for forward compatibility
    """

    archs = set(get_cuda_archs())
    cuda_version = get_cuda_bare_metal_version()
    cc_flags = []

    # Blackwell requires >= 12.8
    if cuda_version is not None and cuda_version >= Version("12.8"):
        if "100" in archs:
            cc_flags += ["-gencode", "arch=compute_100a,code=sm_100a"]

        if "103" in archs:
            cc_flags += ["-gencode", "arch=compute_103a,code=sm_103a"]

        # Thor rename: 12.9 uses sm_101; 13.0+ uses sm_110
        if "110" in archs:
            if cuda_version >= Version("13.0"):
                cc_flags += ["-gencode", "arch=compute_110f,code=sm_110"]
            elif cuda_version >= Version("12.9"):
                # Provide Thor support for CUDA 12.9 via sm_101
                cc_flags += ["-gencode", "arch=compute_101f,code=sm_101"]
            # else: no Thor support in older toolkits

        if "120" in archs:
            # sm_120 is supported in CUDA 12.8/12.9+ toolkits
            if cuda_version >= Version("12.9"):
                cc_flags += ["-gencode", "arch=compute_120f,code=sm_120"]
            else:
                cc_flags += ["-gencode", "arch=compute_120a,code=sm_120a"]

    return cc_flags


def get_platform() -> str:
    if sys.platform.startswith("linux"):
        return f"linux_{platform.uname().machine}"
    if sys.platform == "darwin":
        mac_version = ".".join(platform.mac_ver()[0].split(".")[:2])
        return f"macosx_{mac_version}_x86_64"
    if sys.platform == "win32":
        return "win_amd64"

    msg = f"Unsupported platform: {sys.platform}"
    raise ValueError(msg)


def get_wheel_url() -> tuple[str, str]:
    torch_version_raw = parse(torch.__version__)
    python_version = f"cp{sys.version_info.major}{sys.version_info.minor}"
    platform_name = get_platform()
    torch_version = f"{torch_version_raw.major}.{torch_version_raw.minor}"
    cxx11_abi = str(torch._C._GLIBCXX_USE_CXX11_ABI).upper()  # noqa: SLF001

    # We only compile for CUDA 12.8 to save CI time. Minor versions should be
    # compatible.
    torch_cuda_version = parse("12.8")
    cuda_version = f"cu{torch_cuda_version.major}"

    wheel_filename = (
        f"{PACKAGE_NAME}-{PACKAGE_VERSION}+{cuda_version}torch{torch_version}"
        f"cxx11abi{cxx11_abi}-{python_version}-{python_version}-{platform_name}.whl"
    )

    return f"{BASE_WHEEL_URL}/v{PACKAGE_VERSION}/{wheel_filename}", wheel_filename


class CachedWheelsCommand(bdist_wheel):
    """
    Custom bdist wheel command that checks for pre-built wheels on GitHub Releases.

    The CachedWheelsCommand plugs into the default bdist wheel, which is ran by pip
    when it cannot find an existing wheel (which is currently the case for all
    fouroversix installs). We use the environment parameters to detect whether there is
    already a pre-built version of a compatible wheel available and short-circuits the
    standard full build pipeline.

    Credit: https://github.com/Dao-AILab/flash-attention/blob/main/setup.py
    """

    def run(self) -> None:
        """Run the command."""

        if FORCE_BUILD:
            return super().run()

        wheel_url, wheel_filename = get_wheel_url()
        print(f"Guessing wheel URL: {wheel_url}")

        try:
            urllib.request.urlretrieve(wheel_url, wheel_filename)  # noqa: S310

            # Make the archive
            # Lifted from the root wheel processing command
            # https://github.com/pypa/wheel/blob/cf71108ff9f6ffc36978069acb28824b44ae028e/src/wheel/bdist_wheel.py#LL381C9-L381C85
            if not Path(self.dist_dir).exists():
                Path(self.dist_dir).mkdir(parents=True, exist_ok=True)

            impl_tag, abi_tag, plat_tag = self.get_tag()
            archive_basename = f"{self.wheel_dist_name}-{impl_tag}-{abi_tag}-{plat_tag}"

            wheel_path = Path(self.dist_dir) / (archive_basename + ".whl")
            print(f"Raw wheel path: {wheel_path}")
            Path(wheel_filename).rename(wheel_path)
        except (urllib.error.HTTPError, urllib.error.URLError):
            print("Precompiled wheel not found. Building from source...")
            # If the wheel could not be downloaded, build from source
            super().run()


class NinjaBuildExtension(BuildExtension):
    """
    Custom build extension that tells Ninja how many jobs to run.

    Credit: https://github.com/Dao-AILab/flash-attention/blob/main/setup.py
    """

    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]) -> None:
        # do not override env MAX_JOBS if already exists
        if not os.environ.get("MAX_JOBS"):
            try:
                import psutil

                # calculate the maximum allowed NUM_JOBS based on cores
                max_num_jobs_cores = max(1, os.cpu_count() // 2)

                # calculate the maximum allowed NUM_JOBS based on free memory
                free_memory_gb = psutil.virtual_memory().available / (
                    1024**3
                )  # free memory in GB
                max_num_jobs_memory = int(
                    free_memory_gb / 9,
                )  # each JOB peak memory cost is ~8-9GB when threads = 4

                # pick lower value of jobs based on cores vs memory metric to minimize
                # oom and swap usage during compilation
                max_jobs = max(1, min(max_num_jobs_cores, max_num_jobs_memory))
                os.environ["MAX_JOBS"] = str(max_jobs)
            except ImportError:
                warnings.warn(
                    "psutil not found, install psutil and ninja to get better build "
                    "performance",
                    stacklevel=1,
                )

        super().__init__(*args, **kwargs)


if SKIP_CUDA_BUILD:
    warnings.warn(
        "SKIP_CUDA_BUILD is set to 1, installing fouroversix without quantization and "
        "matmul kernels",
        stacklevel=1,
    )

    ext_modules = None
else:
    if Path(".git").exists():
        subprocess.run(
            [
                "git",
                "submodule",
                "update",
                "--init",
                "third_party/cutlass",
            ],
            check=True,
        )
    elif not Path("third_party/cutlass").exists():
        msg = (
            "third_party/cutlass is missing, please use source distribution or git "
            "clone"
        )
        raise RuntimeError(msg)

    # The compiler flag -D_GLIBCXX_USE_CXX11_ABI is set to be the same as
    # torch._C._GLIBCXX_USE_CXX11_ABI
    # https://github.com/pytorch/pytorch/blob/8472c24e3b5b60150096486616d98b7bea01500b/torch/utils/cpp_extension.py#L920
    if FORCE_CXX11_ABI:
        torch._C._GLIBCXX_USE_CXX11_ABI = True  # noqa: SLF001

    setup_dir = Path(__file__).parent
    kernels_dir = setup_dir / "src" / "fouroversix" / "csrc"
    sources = [
        path.relative_to(Path(__file__).parent).as_posix()
        for ext in ["**/*.cu", "**/*.cpp"]
        for path in kernels_dir.glob(ext)
    ]

    cxx_compile_args = ["-std=c++17"]
    nvcc_compile_args = [
        "-std=c++17",
        "--expt-relaxed-constexpr",
        "-Xcompiler",
        "-funroll-loops",
        "-Xcompiler",
        "-finline-functions",
        *get_cuda_gencodes(),
    ]

    if CUTLASS_DEBUG:
        nvcc_compile_args.extend(
            [
                "-O0",
                "-DCUTLASS_DEBUG_TRACE_LEVEL=3",
                "-DCUTLASS_DEBUG_ENABLE=1",
                "-g",
            ],
        )
    else:
        cxx_compile_args.extend(["-O3"])
        nvcc_compile_args.extend(["-O3", "-DNDEBUG"])

    ext_modules = [
        CUDAExtension(
            "fouroversix._C",
            sources,
            extra_compile_args={"cxx": cxx_compile_args, "nvcc": nvcc_compile_args},
            include_dirs=[
                setup_dir / "third_party/cutlass/examples/common",
                setup_dir / "third_party/cutlass/include",
                setup_dir / "third_party/cutlass/tools/util/include",
                kernels_dir / "include",
            ],
        ),
    ]

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    ext_modules=ext_modules,
    cmdclass={
        "bdist_wheel": CachedWheelsCommand,
        "build_ext": NinjaBuildExtension,
    },
    include_package_data=True,
)



from .conversions import WeightConversions
from .gpt_oss import FourOverSixGptOssDeserialize, GptOssWeightConverter

__all__ = ["FourOverSixGptOssDeserialize", "GptOssWeightConverter", "WeightConversions"]
